from __future__ import annotations

from typing import Optional

from heudiconv.utils import SeqInfo


def create_key(
    template: Optional[str],
    outtype: tuple[str, ...] = ("nii.gz", "dicom"),
    annotation_classes: None = None,
) -> tuple[str, tuple[str, ...], None]:
    if template is None or not template:
        raise ValueError("Template must be a valid format string")
    return (template, outtype, annotation_classes)


def infotodict(seqinfo: list[SeqInfo]) -> dict[tuple[str, tuple[str, ...], None], list]:
    """Heuristic evaluator for determining which runs belong where

    allowed template fields - follow python string module:

    item: index within category
    subject: participant id
    seqitem: run number during scanning
    subindex: sub index within group
    """
    t1 = create_key("sub-{subject}/anat/sub-{subject}_T1w")
    t2 = create_key("sub-{subject}/anat/sub-{subject}_T2w")
    rest = create_key("sub-{subject}/func/sub-{subject}_task-rest_run-{item:02d}_bold")
    rest_sbref = create_key(
        "sub-{subject}/func/sub-{subject}_task-rest_run-{item:02d}_sbref"
    )
    face = create_key("sub-{subject}/func/sub-{subject}_task-face_run-{item:02d}_bold")
    face_sbref = create_key(
        "sub-{subject}/func/sub-{subject}_task-face_run-{item:02d}_sbref"
    )
    gamble = create_key(
        "sub-{subject}/func/sub-{subject}_task-gambling_run-{item:02d}_bold"
    )
    gamble_sbref = create_key(
        "sub-{subject}/func/sub-{subject}_task-gambling_run-{item:02d}_sbref"
    )
    conflict = create_key(
        "sub-{subject}/func/sub-{subject}_task-conflict_run-{item:02d}_bold"
    )
    conflict_sbref = create_key(
        "sub-{subject}/func/sub-{subject}_task-conflict_run-{item:02d}_sbref"
    )
    dwi = create_key("sub-{subject}/dwi/sub-{subject}_run-{item:02d}_dwi")
    dwi_sbref = create_key("sub-{subject}/dwi/sub-{subject}_run-{item:02d}_sbref")
    fmap = create_key("sub-{subject}/fmap/sub-{subject}_dir-{dir}_run-{item:02d}_epi")

    info: dict[tuple[str, tuple[str, ...], None], list] = {
        t1: [],
        t2: [],
        rest: [],
        face: [],
        gamble: [],
        conflict: [],
        dwi: [],
        rest_sbref: [],
        face_sbref: [],
        gamble_sbref: [],
        conflict_sbref: [],
        dwi_sbref: [],
        fmap: [],
    }

    for s in seqinfo:
        # T1 and T2 scans
        if (s.dim3 == 208) and (s.dim4 == 1) and ("T1w" in s.protocol_name):
            info[t1] = [s.series_id]
        if (s.dim3 == 208) and ("T2w" in s.protocol_name):
            info[t2] = [s.series_id]
        # diffusion scans
        if "dMRI_dir9" in s.protocol_name:
            key = None
            if s.dim4 >= 99:
                key = dwi
            elif (s.dim4 == 1) and ("SBRef" in s.series_description):
                key = dwi_sbref
            if key:
                info[key].append({"item": s.series_id})
        # functional scans
        if "fMRI" in s.protocol_name:
            tasktype = s.protocol_name.split("fMRI")[1].split("_")[1]
            key = None
            if s.dim4 in [420, 215, 338, 280]:
                if "rest" in tasktype:
                    key = rest
                if "face" in tasktype:
                    key = face
                if "conflict" in tasktype:
                    key = conflict
                if "gambling" in tasktype:
                    key = gamble
            if (s.dim4 == 1) and ("SBRef" in s.series_description):
                if "rest" in tasktype:
                    key = rest_sbref
                if "face" in tasktype:
                    key = face_sbref
                if "conflict" in tasktype:
                    key = conflict_sbref
                if "gambling" in tasktype:
                    key = gamble_sbref
            if key:
                info[key].append({"item": s.series_id})
        if (s.dim4 == 3) and ("SpinEchoFieldMap" in s.protocol_name):
            dirtype = s.protocol_name.split("_")[-1]
            info[fmap].append({"item": s.series_id, "dir": dirtype})

    # You can even put checks in place for your protocol
    msg = []
    if len(info[t1]) != 1:
        msg.append("Missing correct number of t1 runs")
    if len(info[t2]) != 1:
        msg.append("Missing correct number of t2 runs")
    if len(info[dwi]) != 4:
        msg.append("Missing correct number of dwi runs")
    if len(info[rest]) != 4:
        msg.append("Missing correct number of resting runs")
    if len(info[face]) != 2:
        msg.append("Missing correct number of faceMatching runs")
    if len(info[conflict]) != 4:
        msg.append("Missing correct number of conflict runs")
    if len(info[gamble]) != 2:
        msg.append("Missing correct number of gamble runs")
    if msg:
        raise ValueError("\n".join(msg))
    return info



from __future__ import annotations

from typing import Optional

import pydicom as dcm

from heudiconv.utils import SeqInfo

scaninfo_suffix = ".json"


def create_key(
    template: Optional[str],
    outtype: tuple[str, ...] = ("nii.gz",),
    annotation_classes: None = None,
) -> tuple[str, tuple[str, ...], None]:
    if template is None or not template:
        raise ValueError("Template must be a valid format string")
    return (template, outtype, annotation_classes)


def filter_dicom(dcmdata: dcm.dataset.Dataset) -> bool:
    """Return True if a DICOM dataset should be filtered out, else False"""
    comments = getattr(dcmdata, "ImageComments", "")
    if len(comments):
        if "reference volume" in comments.lower():
            print("Filter out image with comment '%s'" % comments)
            return True
    return False


def extract_moco_params(
    basename: str, _outypes: tuple[str, ...], dicoms: list[str]
) -> None:
    if "_rec-dico" not in basename:
        return
    from pydicom import dcmread

    # get acquisition time for all dicoms
    dcm_times = [
        (d, float(dcmread(d, stop_before_pixels=True).AcquisitionTime)) for d in dicoms
    ]
    # store MoCo info from image comments sorted by acquisition time
    moco = [
        "\t".join(
            [
                str(float(i))
                for i in dcmread(fn, stop_before_pixels=True)
                .ImageComments.split()[1]
                .split(",")
            ]
        )
        for fn, t in sorted(dcm_times, key=lambda x: x[1])
    ]
    outname = basename[:-4] + "recording-motion_physio.tsv"
    with open(outname, "wt") as fp:
        for m in moco:
            fp.write("%s\n" % (m,))


custom_callable = extract_moco_params


def infotodict(
    seqinfo: list[SeqInfo],
) -> dict[tuple[str, tuple[str, ...], None], list[str]]:
    """Heuristic evaluator for determining which runs belong where

    allowed template fields - follow python string module:

    item: index within category
    subject: participant id
    seqitem: run number during scanning
    subindex: sub index within group
    """

    info: dict[tuple[str, tuple[str, ...], None], list[str]] = {}
    for s in seqinfo:
        if "_bold_" not in s.protocol_name:
            continue
        if "_coverage" not in s.protocol_name:
            label = "orientation%s_run-{item:02d}"
        else:
            label = "coverage%s"
        resolution = s.protocol_name.split("_")[5][:-3]
        assert float(resolution)
        if s.is_motion_corrected:
            label = label % ("_rec-dico",)
        else:
            label = label % ("",)

        templ = "ses-%smm/func/{subject}_ses-%smm_task-%s_bold" % (
            resolution,
            resolution,
            label,
        )

        key = create_key(templ)

        if key not in info:
            info[key] = []
        info[key].append(s.series_id)

    return info



#!/usr/bin/env python
from __future__ import annotations

import argparse
from collections import OrderedDict
import json
import logging
import os
import os.path as op
from pathlib import Path
import re
import shlex
import subprocess
import time
from typing import Any, Optional

import inotify.adapters
from inotify.constants import IN_CREATE, IN_ISDIR, IN_MODIFY
from tinydb import TinyDB

_DEFAULT_LOG_FORMAT = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
_LOGGER = logging.getLogger(__name__)

MASK = IN_MODIFY | IN_CREATE
MASK_NEWDIR = IN_CREATE | IN_ISDIR
WAIT_TIME = 86400  # in seconds


# def _configure_logging():
_LOGGER.setLevel(logging.DEBUG)
ch = logging.StreamHandler()
formatter = logging.Formatter(_DEFAULT_LOG_FORMAT)
ch.setFormatter(formatter)
_LOGGER.addHandler(ch)


def run_heudiconv(args: list[str]) -> tuple[str, dict[str, Any]]:
    info_dict: dict[str, Any] = dict()
    proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    cmd = " ".join(map(shlex.quote, args))
    return_code = proc.wait()
    if return_code == 0:
        _LOGGER.info("Done running %s", cmd)
        info_dict["success"] = 1
    else:
        _LOGGER.error("%s failed", cmd)
        info_dict["success"] = 0
    # get info on what we run
    stdout = proc.communicate()[0].decode("utf-8")
    match = re.match("INFO: PROCESSING STARTS: (.*)", stdout)
    info_dict_ = json.loads(match.group(1) if match else "{}")
    info_dict.update(info_dict_)
    return stdout, info_dict


def process(
    paths2process: dict[str, float],
    db: Optional[TinyDB],
    wait: int | float = WAIT_TIME,
    logdir: str = "log",
) -> None:
    # if paths2process and
    # time.time() - os.path.getmtime(paths2process[0]) > WAIT_TIME:
    processed: list[str] = []
    for path, mod_time in paths2process.items():
        if time.time() - mod_time > wait:
            # process_me = paths2process.popleft().decode('utf-8')
            process_me = path
            process_dict: dict[str, Any] = {
                "input_path": process_me,
                "accession_number": op.basename(process_me),
            }
            print("Time to process {0}".format(process_me))
            stdout, run_dict = run_heudiconv(["ls", "-l", process_me])
            process_dict.update(run_dict)
            if db is not None:
                db.insert(process_dict)
            # save log
            log = Path(logdir, process_dict["accession_number"] + ".log")
            log.write_text(stdout)
            # if we processed it, or it failed,
            # we need to remove it to avoid running it again
            processed.append(path)
    for processed_path in processed:
        del paths2process[processed_path]


def monitor(
    topdir: str = "/tmp/new_dir",
    check_ptrn: str = "/20../../..",
    db: Optional[TinyDB] = None,
    wait: int | float = WAIT_TIME,
    logdir: str = "log",
) -> None:
    # make logdir if not existent
    try:
        os.makedirs(logdir)
    except OSError:
        pass
    # paths2process = deque()
    paths2process: dict[str, float] = OrderedDict()
    # watch only today's folder
    path_re = re.compile("(%s%s)/?$" % (topdir, check_ptrn))
    i = inotify.adapters.InotifyTree(topdir.encode())  # , mask=MASK)
    for event in i.event_gen():
        if event is not None:
            (header, type_names, watch_path, filename) = event
            _LOGGER.info(
                "WD=(%d) MASK=(%d) COOKIE=(%d) LEN=(%d) MASK->NAMES=%s"
                " WATCH-PATH=[%s] FILENAME=[%s]",
                header.wd,
                header.mask,
                header.cookie,
                header.len,
                type_names,
                watch_path.decode("utf-8"),
                filename.decode("utf-8"),
            )
            if header.mask == MASK_NEWDIR and path_re.match(watch_path.decode("utf-8")):
                # we got our directory, now let's do something on it
                newpath2process = op.join(watch_path, filename).decode("utf-8")
                # paths2process.append(newpath2process)
                # update time
                paths2process[newpath2process] = time.time()
                print(newpath2process, time.time())
            # check if we need to update the time
            for path in paths2process.keys():
                if path in watch_path.decode("utf-8"):
                    paths2process[path] = time.time()
                    print("Updating {0}: {1}".format(path, paths2process[path]))

        # check if there's anything to process
        process(paths2process, db, wait=wait, logdir=logdir)


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        prog="monitor.py",
        description=(
            "Small monitoring script to detect new directories and " "process them"
        ),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("path", help="Which directory to monitor")
    parser.add_argument(
        "--check_ptrn",
        "-p",
        help="regexp pattern for which subdirectories to check",
        default="/20../../..",
    )
    parser.add_argument(
        "--database", "-d", help="database location", default="database.json"
    )
    parser.add_argument(
        "--wait_time",
        "-w",
        help="After how long should we start processing datasets? (in seconds)",
        default=86400,
        type=float,
    )
    parser.add_argument(
        "--logdir", "-l", help="Where should we save the logs?", default="log"
    )
    return parser.parse_args()


def main() -> None:
    parsed = parse_args()
    print("Got {0}".format(parsed))
    # open database
    db = TinyDB(parsed.database, default_table="heudiconv")
    monitor(
        parsed.path, parsed.check_ptrn, db, wait=parsed.wait_time, logdir=parsed.logdir
    )


if __name__ == "__main__":
    main()



from __future__ import annotations

from glob import glob
import os
import os.path as op
from pathlib import Path
import shutil
import stat

import pytest

from .utils import TESTS_DATA_PATH
from ..parser import get_extracted_dicoms


def _get_dicoms_archive(tmpdir: Path, fmt: str) -> list[str]:
    tmp_file = tmpdir / "dicom"
    archive = shutil.make_archive(
        str(tmp_file), format=fmt, root_dir=TESTS_DATA_PATH, base_dir="01-anat-scout"
    )
    return [archive]


@pytest.fixture
def get_dicoms_archive(tmpdir: Path, request: pytest.FixtureRequest) -> list[str]:
    return _get_dicoms_archive(tmpdir, fmt=request.param)


@pytest.fixture
def get_dicoms_gztar(tmpdir: Path) -> list[str]:
    return _get_dicoms_archive(tmpdir, "gztar")


@pytest.fixture
def get_dicoms_list() -> list[str]:
    return glob(op.join(TESTS_DATA_PATH, "01-anat-scout", "*"))


def test_get_extracted_dicoms_single_session_is_none(
    get_dicoms_gztar: list[str],
) -> None:
    for session_, _ in get_extracted_dicoms(get_dicoms_gztar):
        assert session_ is None


def test_get_extracted_dicoms_multple_session_integers(
    get_dicoms_gztar: list[str],
) -> None:
    sessions = [
        session
        for session, _ in get_extracted_dicoms(get_dicoms_gztar + get_dicoms_gztar)
    ]

    assert sessions == ["0", "1"]


@pytest.mark.parametrize(
    "get_dicoms_archive", ("tar", "gztar", "zip", "bztar", "xztar"), indirect=True
)
def test_get_extracted_dicoms_from_archives(get_dicoms_archive: list[str]) -> None:
    for _, files in get_extracted_dicoms(get_dicoms_archive):
        # check that the only file is the one called "0001.dcm"
        endswith = all(file.endswith("0001.dcm") for file in files)

        # check that permissions were set
        mode = all(stat.S_IMODE(os.stat(file).st_mode) == 448 for file in files)

        # check for absolute paths
        absolute = all(op.isabs(file) for file in files)

        assert all([endswith, mode, absolute])


def test_get_extracted_dicoms_from_file_list(get_dicoms_list: list[str]) -> None:
    for _, files in get_extracted_dicoms(get_dicoms_list):
        assert all(op.isfile(file) for file in files)


def test_get_extracted_dicoms_from_mixed_list(
    get_dicoms_list: list[str], get_dicoms_gztar: list[str]
) -> None:
    for _, files in get_extracted_dicoms(get_dicoms_list + get_dicoms_gztar):
        assert all(op.isfile(file) for file in files)


def test_get_extracted_from_empty_list() -> None:
    assert not len(get_extracted_dicoms([]))



"""
(AKA dbic-bids) Flexible heuristic to establish BIDS DataLad datasets hierarchy

Initially developed and deployed at Dartmouth Brain Imaging Center
(http://dbic.dartmouth.edu) using Siemens Prisma 3T under the umbrellas of the
Center of Reproducible Neuroimaging Computation (ReproNim, http://repronim.org)
and Center for Open Neuroscience (CON, http://centerforopenneuroscience.org).

## Dataset ownership/location

Datasets will be arranged in a hierarchy similar to how study/exam cards are
arranged at the scanner console.  You should have

- "region" defined per each PI,
 - on the first level most probably as PI_StudentOrRA/ (e.g., Gobbini_Matteo)
   - StudyID_StudyName/   (e.g. 1002_faceangles)
     - Arbitrary name for the exam card -- it doesn't get into Study Description.

Selecting specific exam card would populate Study Description field using
aforementioned levels, which will be used by this heuristic to decide on the
location of the dataset.

In case of multiple sessions, it is recommended to generate separate "cards"
per each session.

## Sequence naming on the scanner console

Sequence names on the scanner must follow this specification to avoid manual
conversion/handling:

  [PREFIX:][WIP ]<datatype[-<suffix>]>[_ses-<SESID>][_task-<TASKID>][_acq-<ACQLABEL>][_run-<RUNID>][_dir-<DIR>][<more BIDS>][__<custom>]

where
 [PREFIX:] - leading capital letters followed by : are stripped/ignored
 [WIP ] - prefix is stripped/ignored (added by Philips for patch sequences)
 <...> - value to be entered
 [...] - optional -- might be nearly mandatory for some modalities (e.g.,
         run for functional) and very optional for others
 *ID - alpha-numerical identifier (e.g. 01,02, pre, post, pre01) for a run,
       task, session. Note that makes more sense to use numerical values for
       RUNID (e.g., _run-01, _run-02) for obvious sorting and possibly
       descriptive ones for e.g. SESID (_ses-movie, _ses-localizer)


<datatype[-suffix]>
   a known BIDS sequence datatype which is usually a name of the folder under
   subject's directory. And (optional) suffix is a specific sequence type
   (e.g., "bold" for func, or "T1w" for "anat"), which could often
   (but not always) be deduced from DICOM. Known to ReproIn BIDS modalities
   are:

     anat - anatomical data.  Might also be collected multiple times across
            runs (e.g. if subject is taken out of magnet etc), so could
            (optionally) have "_run" definition attached. For "standard anat"
            suffixes, please consult to "8.3 Anatomy imaging data" but most
            common are 'T1w', 'T2w', 'angio'.
     beh  - behavioral data. known but not "treated".
     func - functional (AKA task, including resting state) data.
            Typically contains multiple runs, and might have multiple different
            tasks different per each run
            (e.g. _task-memory_run-01, _task-oddball_run-02)
     fmap - field maps
     dwi  - diffusion weighted imaging (also can as well have runs)

   The other BIDS modalities are not known ATM and their data will not be
   converted and will be just skipped (with a warning). Full list of datatypes
   can be found at
   https://github.com/bids-standard/bids-specification/blob/v1.7.0/src/schema/objects/datatypes.yaml
   and their corresponding suffixes at
   https://github.com/bids-standard/bids-specification/tree/v1.7.0/src/schema/rules/datatypes

_ses-<SESID> (optional)
    a session.  Having a single sequence within a study would make that study
    follow "multi-session" layout. A common practice to have a _ses specifier
    within the scout sequence name. You can either specify explicit session
    identifier (SESID) or just say to maintain, create (starts with 1).
    You can also use _ses-{date} in case of scanning phantoms or non-human
    subjects and wanting sessions to be coded by the acquisition date.

_task-<TASKID> (optional)
    a short name for a task performed during that run.  If not provided and it
    is a func sequence, _task-UNKNOWN will be automatically added to comply with
    BIDS. Consult http://www.cognitiveatlas.org/tasks on known tasks.

_acq-<ACQLABEL> (optional)
    a short custom label to distinguish a different set of parameters used for
    acquiring the same modality (e.g. _acq-highres, _acq-lowres  etc)

_run-<RUNID> (optional)
    a (typically functional) run. The same idea as with SESID.

_dir-[AP,PA,LR,RL,VD,DV] (optional)
    to be used for fmap images, whenever a pair of the SE images is collected
    to be used to estimate the fieldmap

<more BIDS> (optional)
    any other fields (e.g. _acq-) from BIDS acquisition

__<custom> (optional)
  after two underscores any arbitrary comment which will not matter to how
  layout in BIDS. But that one theoretically should not be necessary,
  and (ab)use of it would just signal lack of thought while preparing sequence
  name to start with since everything could have been expressed in BIDS fields.

## Last moment checks/FAQ:

- Functional runs should have _task-<TASKID> field defined
- Do not use "+", "_" or "-" within SESID, TASKID, ACQLABEL, RUNID,  so we
  could detect "canceled" runs.
- If run was canceled -- just copy canceled run (with the same index) and re-run
  it. Files with overlapping name will be considered duplicate/canceled session
  and only the last one would remain.  The others would acquire
  __dup0<number>  suffix.

Although we still support "-" and "+" used within SESID and TASKID, their use is
not recommended, thus not listed here

## Scanner specifics

We perform following actions regardless of the type of scanner, but applied
generally to accommodate limitations imposed by different manufacturers/models:

### Philips

- We replace all ( with { and ) with } to be able e.g. to specify session {date}
- "WIP " prefix unconditionally added by the scanner is stripped
"""

from __future__ import annotations

from collections.abc import Iterable
from glob import glob
import hashlib
import logging
import os.path
import re
from typing import Any, Optional, TypeVar

import pydicom as dcm

from heudiconv.due import Doi, due
from heudiconv.utils import SeqInfo, StudySessionInfo

lgr = logging.getLogger("heudiconv")

T = TypeVar("T")

# Terminology to harmonise and use to name variables etc
# experiment
#  subject
#   [session]
#    exam (AKA scanning session) - currently seqinfo, unless brought together from multiple
#     series  (AKA protocol?)
#      - series_spec - deduced from fields the spec (literal value)
#      - series_info - the dictionary with fields parsed from series_spec

# Which fields in seqinfo (in this order) to check for the ReproIn spec
series_spec_fields = ("protocol_name", "series_description")

# dictionary from accession-number to runs that need to be marked as bad
# NOTE: even if filename has number that is 0-padded, internally no padding
# is done
fix_accession2run: dict[str, list[str]] = {
    # e.g.:
    # 'A000035': ['^8-', '^9-'],
}

# A dictionary containing fixes/remapping for sequence names per study.
# Keys are md5sum of study_description from DICOMs, in the form of PI-Experimenter^protocolname
# You can use `heudiconv -f reproin --command ls --files  PATH
# to list the "study hash".
# Values are list of tuples in the form (regex_pattern, substitution).
# If the  key is an empty string`''''`, it would apply to any study.
protocols2fix: dict[str | re.Pattern[str], list[tuple[str, str]]] = {
    # e.g., QA:
    # '43b67d9139e8c7274578b7451ab21123':
    #     [
    #      ('BOLD_p2_s4_3\.5mm', 'func_task-rest_acq-p2-s4-3.5mm'),
    #      ('BOLD_', 'func_task-rest'),
    #      ('_p2_s4',        '_acq-p2-s4'),
    #      ('_p2', '_acq-p2'),
    #     ],
    # '':  # for any study example with regexes used
    #     [
    #         ('AAHead_Scout_.*', 'anat-scout'),
    #         ('^dti_.*', 'dwi'),
    #         ('^.*_distortion_corr.*_([ap]+)_([12])', r'fmap-epi_dir-\1_run-\2'),
    #         ('^(.+)_ap.*_r(0[0-9])', r'func_task-\1_run-\2'),
    #         ('^t1w_.*', 'anat-T1w'),
    #         # problematic case -- multiple identically named pepolar fieldmap runs
    #         # I guess we will just sacrifice ability to detect canceled runs here.
    #         # And we cannot just use _run+ since it would increment independently
    #         # for ap and then for pa.  We will rely on having ap preceding pa.
    #         # Added  _acq-mb8  so they match the one in funcs
    #         ('func_task-discorr_acq-ap', r'fmap-epi_dir-ap_acq-mb8_run+'),
    #         ('func_task-discorr_acq-pa', r'fmap-epi_dir-pa_acq-mb8_run='),
    # ]
}

# list containing StudyInstanceUID to skip -- hopefully doesn't happen too often
dicoms2skip: list[str] = [
    # e.g.
    # '1.3.12.2.1107.5.2.43.66112.30000016110117002435700000001',
]

DEFAULT_FIELDS = {
    # Let it just be in each json file extracted
    "Acknowledgements": "We thank Terry Sacket and the rest of the DBIC (Dartmouth Brain Imaging "
    "Center) personnel for assistance in data collection, and "
    "Yaroslav O. Halchenko for preparing BIDS dataset. "
    "TODO: adjust to your case.",
}

POPULATE_INTENDED_FOR_OPTS = {
    "matching_parameters": ["ImagingVolume", "Shims"],
    "criterion": "Closest",
}


KNOWN_DATATYPES = {"anat", "func", "dwi", "behav", "fmap"}


def _delete_chars(from_str: str, deletechars: str) -> str:
    return from_str.translate(str.maketrans("", "", deletechars))


def filter_dicom(dcmdata: dcm.dataset.Dataset) -> bool:
    """Return True if a DICOM dataset should be filtered out, else False"""
    return True if dcmdata.StudyInstanceUID in dicoms2skip else False


def filter_files(_fn: str) -> bool:
    """Return True if a file should be kept, else False.

    ATM reproin does not do any filtering. Override if you need to add some
    """
    return True


def create_key(
    subdir: Optional[str],
    file_suffix: str,
    outtype: tuple[str, ...] = ("nii.gz", "dicom"),
    annotation_classes: None = None,
    prefix: str = "",
) -> tuple[str, tuple[str, ...], None]:
    if not subdir:
        raise ValueError("subdir must be a valid format string")
    # may be even add "performing physician" if defined??
    template = os.path.join(
        prefix,
        "{bids_subject_session_dir}",
        subdir,
        "{bids_subject_session_prefix}_%s" % file_suffix,
    )
    return template, outtype, annotation_classes


def md5sum(string: Optional[str]) -> str:
    """Computes md5sum of a string"""
    if not string:
        return ""  # not None so None was not compared to strings
    m = hashlib.md5(string.encode())
    return m.hexdigest()


def get_study_description(seqinfo: list[SeqInfo]) -> str:
    # Centralized so we could fix/override
    v = get_unique(seqinfo, "study_description")
    assert isinstance(v, str)
    return v


def get_study_hash(seqinfo: list[SeqInfo]) -> str:
    # XXX: ad hoc hack
    return md5sum(get_study_description(seqinfo))


def fix_canceled_runs(seqinfo: list[SeqInfo]) -> list[SeqInfo]:
    """Function that adds cancelme_ to known bad runs which were forgotten"""
    if not fix_accession2run:
        return seqinfo  # nothing to do
    for i, s in enumerate(seqinfo):
        accession_number = s.accession_number
        if accession_number and accession_number in fix_accession2run:
            lgr.info(
                "Considering some runs possibly marked to be "
                "canceled for accession %s",
                accession_number,
            )
            # This code is reminiscent of prior logic when operating on
            # a single accession, but left as is for now
            badruns = fix_accession2run[accession_number]
            badruns_pattern = "|".join(badruns)
            if re.match(badruns_pattern, s.series_id):
                lgr.info("Fixing bad run {0}".format(s.series_id))
                fixedkwargs = dict()
                for key in series_spec_fields:
                    fixedkwargs[key] = "cancelme_" + getattr(s, key)
                seqinfo[i] = s._replace(**fixedkwargs)
    return seqinfo


def fix_dbic_protocol(seqinfo: list[SeqInfo]) -> list[SeqInfo]:
    """Ad-hoc fixup for existing protocols.

    It will operate in 3 stages on `protocols2fix` records.
    1. consider a record which has md5sum of study_description
    2. apply all substitutions, where key is a regular expression which
       successfully searches (not necessarily matches, so anchor appropriately)
       study_description
    3. apply "catch all" substitutions in the key containing an empty string

    3. is somewhat redundant since `re.compile('.*')` could match any, but is
    kept for simplicity of its specification.
    """

    study_hash = get_study_hash(seqinfo)
    study_description = get_study_description(seqinfo)

    # We will consider first study specific (based on hash)
    if study_hash in protocols2fix:
        _apply_substitutions(
            seqinfo, protocols2fix[study_hash], "study (%s) specific" % study_hash
        )
    # Then go through all regexps returning regex "search" result
    # on study_description
    for sub, substitutions in protocols2fix.items():
        if isinstance(sub, re.Pattern) and sub.search(study_description):
            _apply_substitutions(
                seqinfo, substitutions, "%r regex matching" % sub.pattern
            )
    # and at the end - global
    if "" in protocols2fix:
        _apply_substitutions(seqinfo, protocols2fix[""], "global")

    return seqinfo


def _apply_substitutions(
    seqinfo: list[SeqInfo], substitutions: list[tuple[str, str]], subs_scope: str
) -> None:
    lgr.info("Considering %s substitutions", subs_scope)
    for i, s in enumerate(seqinfo):
        fixed_kwargs = dict()
        # need to replace both protocol_name series_description
        for key in series_spec_fields:
            oldvalue = value = getattr(s, key)
            # replace all I need to replace
            for substring, replacement in substitutions:
                value = re.sub(substring, replacement, value)
            if oldvalue != value:
                lgr.info(" %s: %r -> %r", key, oldvalue, value)
            fixed_kwargs[key] = value
        # namedtuples are immutable
        seqinfo[i] = s._replace(**fixed_kwargs)


def fix_seqinfo(seqinfo: list[SeqInfo]) -> list[SeqInfo]:
    """Just a helper on top of both fixers"""
    # add cancelme to known bad runs
    seqinfo = fix_canceled_runs(seqinfo)
    seqinfo = fix_dbic_protocol(seqinfo)
    return seqinfo


def ls(_study_session: StudySessionInfo, seqinfo: list[SeqInfo]) -> str:
    """Additional ls output for a seqinfo"""
    # assert len(sequences) <= 1  # expecting only a single study here
    # seqinfo = sequences.keys()[0]
    return " study hash: %s" % get_study_hash(seqinfo)


# XXX we killed session indicator!  what should we do now?!!!
# WE DON'T NEED IT -- it will be provided into conversion_info as `session`
# So we just need subdir and file_suffix!
@due.dcite(
    Doi("10.5281/zenodo.1207117"),
    path="heudiconv.heuristics.reproin",
    description="ReproIn heudiconv heuristic for turnkey conversion into BIDS",
)
def infotodict(
    seqinfo: list[SeqInfo],
) -> dict[tuple[str, tuple[str, ...], None], list[str]]:
    """Heuristic evaluator for determining which runs belong where

    allowed template fields - follow python string module:

    item: index within category
    subject: participant id
    seqitem: run number during scanning
    subindex: sub index within group
    session: scan index for longitudinal acq
    """
    seqinfo = fix_seqinfo(seqinfo)
    lgr.info("Processing %d seqinfo entries", len(seqinfo))

    info: dict[tuple[str, tuple[str, ...], None], list[str]] = {}
    skipped: list[str] = []
    skipped_unknown: list[str] = []
    current_run = 0
    run_label: Optional[str] = None  # run-
    dcm_image_iod_spec: Optional[str] = None
    skip_derived = False
    for s in seqinfo:
        # XXX: skip derived sequences, we don't store them to avoid polluting
        # the directory, unless it is the motion corrected ones
        # (will get _rec-moco suffix)
        if skip_derived and s.is_derived and not s.is_motion_corrected:
            skipped.append(s.series_id)
            lgr.debug("Ignoring derived data %s", s.series_id)
            continue

        # possibly apply present formatting in the series_description or protocol name
        for f in "series_description", "protocol_name":
            s = s._replace(**{f: getattr(s, f).format(**s._asdict())})

        template = None
        suffix = ""
        # seq = []

        # figure out type of image from s.image_info -- just for checking ATM
        # since we primarily rely on encoded in the protocol name information
        prev_dcm_image_iod_spec = dcm_image_iod_spec
        if len(s.image_type) > 2:
            # https://dicom.innolitics.com/ciods/cr-image/general-image/00080008
            # 0 - ORIGINAL/DERIVED
            # 1 - PRIMARY/SECONDARY
            # 3 - Image IOD specific specialization (optional)
            dcm_image_iod_spec = s.image_type[2]
            image_type_datatype = {
                # Note: P and M are too generic to make a decision here, could be
                #  for different datatypes (bold, fmap, etc)
                "FMRI": "func",
                "MPR": "anat",
                "DIFFUSION": "dwi",
                "MIP_SAG": "anat",  # angiography
                "MIP_COR": "anat",  # angiography
                "MIP_TRA": "anat",  # angiography
            }.get(dcm_image_iod_spec, None)
        else:
            dcm_image_iod_spec = image_type_datatype = None

        series_info = {}  # For please lintian and its friends
        for sfield in series_spec_fields:
            svalue = getattr(s, sfield)
            series_info = parse_series_spec(svalue)
            if series_info:  # looks like a valid spec - we are done
                series_spec = svalue
                break
            else:
                lgr.debug("Failed to parse reproin spec in .%s=%r", sfield, svalue)

        if not series_info:
            series_spec = None  # we cannot know better
            lgr.warning(
                "Could not determine the series name by looking at " "%s fields",
                ", ".join(series_spec_fields),
            )
            skipped_unknown.append(s.series_id)
            continue

        if dcm_image_iod_spec and dcm_image_iod_spec.startswith("MIP"):
            series_info["acq"] = series_info.get("acq", "") + sanitize_str(
                dcm_image_iod_spec
            )

        datatype = series_info.pop("datatype")
        datatype_suffix = series_info.pop("datatype_suffix", None)

        if image_type_datatype and datatype != image_type_datatype:
            lgr.warning(
                "Deduced datatype to be %s from DICOM, but got %s out of %s",
                image_type_datatype,
                datatype,
                series_spec,
            )

        # if s.is_derived:
        #     # Let's for now stash those close to original images
        #     # TODO: we might want a separate tree for all of this!?
        #     # so more of a parameter to the create_key
        #     #datatype += '/derivative'
        #     # just keep it lower case and without special characters
        #     # XXXX what for???
        #     #seq.append(s.series_description.lower())
        #     prefix = os.path.join('derivatives', 'scanner')
        # else:
        #     prefix = ''
        prefix = ""

        #
        # Figure out the datatype_suffix (BIDS _suffix)
        #
        # If none was provided -- let's deduce it from the information we find:
        # analyze s.protocol_name (series_id is based on it) for full name mapping etc
        if not datatype_suffix:
            if datatype == "func":
                if "_pace_" in series_spec:
                    datatype_suffix = "pace"  # or should it be part of seq-
                elif "P" in s.image_type:
                    datatype_suffix = "phase"
                elif "M" in s.image_type:
                    datatype_suffix = "bold"
                else:
                    # assume bold by default
                    datatype_suffix = "bold"
            elif datatype == "fmap":
                # TODO: support phase1 phase2 like in "Case 2: Two phase images ..."
                if not dcm_image_iod_spec:
                    raise ValueError("Do not know image data type yet to make decision")
                datatype_suffix = {
                    # might want explicit {file_index}  ?
                    # _epi for pepolar fieldmaps, see
                    # https://bids-specification.readthedocs.io/en/stable/04-modality-specific-files/01-magnetic-resonance-imaging-data.html#case-4-multiple-phase-encoded-directions-pepolar
                    "M": "epi" if "dir" in series_info else "magnitude",
                    "P": "phasediff",
                    "DIFFUSION": "epi",  # according to KODI those DWI are the EPIs we need
                }[dcm_image_iod_spec]
            elif datatype == "dwi":
                # label for dwi as well
                datatype_suffix = "dwi"

        #
        # Even if datatype_suffix was provided, for some data we might need to override,
        # since they are complementary files produced along-side with original
        # ones.
        #
        if s.series_description.endswith("_SBRef"):
            datatype_suffix = "sbref"

        if not datatype_suffix:
            # Might be provided by the bids ending within series_spec, we would
            # just want to check if that the last element is not _key-value pair
            bids_ending = series_info.get("bids", None)
            if not bids_ending or "-" in bids_ending.split("_")[-1]:
                lgr.warning(
                    "We ended up with an empty label/suffix for %r", series_spec
                )

        run = series_info.get("run")
        if run is not None:
            # so we have an indicator for a run
            if run == "+":
                # some sequences, e.g.  fmap, would generate two (or more?)
                # sequences -- e.g. one for magnitude(s) and other ones for
                # phases.  In those we must not increment run!
                if dcm_image_iod_spec and dcm_image_iod_spec == "P":
                    if prev_dcm_image_iod_spec != "M":
                        # XXX if we have a known earlier study, we need to always
                        # increase the run counter for phasediff because magnitudes
                        # were not acquired
                        if get_study_hash([s]) == "9d148e2a05f782273f6343507733309d":
                            current_run += 1
                        else:
                            raise RuntimeError(
                                "Was expecting phase image to follow magnitude "
                                "image, but previous one was %r",
                                prev_dcm_image_iod_spec,
                            )
                        # else we do nothing special
                else:  # and otherwise we go to the next run
                    current_run += 1
            elif run == "=":
                if not current_run:
                    current_run = 1
            elif run.isdigit():
                current_run_ = int(run)
                if current_run_ < current_run:
                    lgr.warning(
                        "Previous run (%s) was larger than explicitly specified %s",
                        current_run,
                        current_run_,
                    )
                current_run = current_run_
            else:
                raise ValueError(
                    "Don't know how to deal with run specification %s" % repr(run)
                )
            run_label = "run-%02d" % current_run
        else:
            # if there is no _run -- no run label added
            run_label = None

        # yoh: had a wrong assumption
        # if s.is_motion_corrected:
        #     assert s.is_derived, "Motion corrected images must be 'derived'"

        if s.is_motion_corrected and "rec-" in series_info.get("bids", ""):
            raise NotImplementedError(
                "want to add _rec-moco but there is _rec- already"
            )

        def from_series_info(name: str) -> Optional[str]:
            """A little helper to provide _name-value if series_info knows it

            Returns None otherwise
            """
            if series_info.get(name):  # noqa: B023
                return "%s-%s" % (name, series_info[name])  # noqa: B023
            else:
                return None

        # TODO: get order from schema, do not hardcode. ATM could be checked at
        # https://bids-specification.readthedocs.io/en/stable/99-appendices/04-entity-table.html
        # https://github.com/bids-standard/bids-specification/blob/HEAD/src/schema/rules/entities.yaml
        # ATM we at large rely on possible (re)ordering according to schema to be done
        # by heudiconv, not reproin here.
        filename_suffix_parts = [
            from_series_info("task"),
            from_series_info("acq"),
            # But we want to add an indicator in case it was motion corrected
            # in the magnet. ref sample  /2017/01/03/qa
            None if not s.is_motion_corrected else "rec-moco",
            from_series_info("dir"),
            series_info.get("bids"),
            run_label,
            datatype_suffix,
        ]
        # filter those which are None, and join with _
        suffix = "_".join(filter(bool, filename_suffix_parts))  # type: ignore[arg-type]

        # # .series_description in case of
        # sdesc = s.study_description
        # # temporary aliases for those phantoms which we already collected
        # # so we rename them into this
        # #MAPPING
        #
        # # the idea ias to have sequence names in the format like
        # # bids_<subdir>_bidsrecord
        # # in bids record we could have  _run[+=]
        # #  which would say to either increment run number from already encountered
        # #  or reuse the last one
        # if seq:
        #     suffix += 'seq-%s' % ('+'.join(seq))

        # For scouts -- we want only dicoms
        # https://github.com/nipy/heudiconv/issues/145
        outtype: tuple[str, ...]
        if (
            "_Scout" in s.series_description
            or (
                datatype == "anat"
                and datatype_suffix
                and datatype_suffix.startswith("scout")
            )
            or (s.series_description.lower() == s.protocol_name.lower() + "_setter")
        ):
            outtype = ("dicom",)
        else:
            outtype = ("nii.gz", "dicom")

        template = create_key(datatype, suffix, prefix=prefix, outtype=outtype)
        # we wanted ordered dict for consistent demarcation of dups
        if template not in info:
            info[template] = []
        info[template].append(s.series_id)

    if skipped:
        lgr.info("Skipped %d sequences: %s" % (len(skipped), skipped))
    if skipped_unknown:
        lgr.warning(
            "Could not figure out where to stick %d sequences: %s"
            % (len(skipped_unknown), skipped_unknown)
        )

    info = get_dups_marked(info)  # mark duplicate ones with __dup-0x suffix
    return info


def get_dups_marked(
    info: dict[tuple[str, tuple[str, ...], None], list[T]], per_series: bool = True
) -> dict[tuple[str, tuple[str, ...], None], list[T]]:
    """

    Parameters
    ----------
    info
    per_series: bool
      If set to False, it would create growing index through all series. That
      could lead to non-desired effects if some "multi file" scans (such as
      fmap with magnitude{1,2} and phasediff) would not be able to associate
      multiple files for the same acquisition.   By default (True) dup indices
      would be per each series (change introduced in 0.5.2)

    Returns
    -------

    """
    # analyze for "cancelled" runs, if run number was explicitly specified and
    # thus we ended up with multiple entries which would mean that older ones
    #  were "cancelled"
    info = info.copy()
    dup_id = 0
    for template, series_ids in list(info.items()):
        if len(series_ids) > 1:
            lgr.warning(
                "Detected %d duplicated run(s) for template %s: %s",
                len(series_ids) - 1,
                template[0],
                series_ids[:-1],
            )
            # copy the duplicate ones into separate ones
            if per_series:
                dup_id = 0  # reset since declared per series
            for dup_series_id in series_ids[:-1]:
                dup_id += 1
                dup_template = ("%s__dup-%02d" % (template[0], dup_id),) + template[1:]
                # There must have not been such a beast before!
                if dup_template in info:
                    raise AssertionError(
                        "{} is already known to info={}. "
                        "May be a bug for per_series=True handling?"
                        "".format(dup_template, info)
                    )
                info[dup_template] = [dup_series_id]
            info[template] = series_ids[-1:]
        assert len(info[template]) == 1
    return info


def get_unique(seqinfos: list[SeqInfo], attr: str) -> Any:
    """Given a list of seqinfos, which must have come from a single study,
    get specific attr, which must be unique across all of the entries

    If not -- fail!

    """
    values = set(getattr(si, attr) for si in seqinfos)
    if len(values) != 1:
        raise AssertionError(
            f"Was expecting a single value for attribute {attr!r} "
            f"but got: {', '.join(sorted(values))}"
        )
    return values.pop()


# TODO: might need to do grouping per each session and return here multiple
# hits, or may be we could just somehow demarkate that it will be multisession
# one and so then later value parsed (again) in infotodict would be used???
def infotoids(seqinfos: Iterable[SeqInfo], outdir: str) -> dict[str, Optional[str]]:
    seqinfo_lst = list(seqinfos)
    # decide on subjid and session based on patient_id
    lgr.info("Processing sequence infos to deduce study/session")
    study_description = get_study_description(seqinfo_lst)
    study_description_hash = md5sum(study_description)
    subject = fixup_subjectid(get_unique(seqinfo_lst, "patient_id"))
    # TODO:  fix up subject id if missing some 0s
    if study_description:
        # Generally it is a ^ but if entered manually, ppl place space in it
        split = re.split("[ ^]", study_description, maxsplit=1)
        # split first one even more, since could be PI_Student or PI-Student
        split = re.split("[-_]", split[0], maxsplit=1) + split[1:]

        # locator = study_description.replace('^', '/')
        locator = "/".join(split)
    else:
        locator = "unknown"

    # TODO: actually check if given study is study we would care about
    # and if not -- we should throw some ???? exception

    # So -- use `outdir` and locator etc to see if for a given locator/subject
    # and possible ses+ in the sequence names, so we would provide a sequence
    # So might need to go through  parse_series_spec(s.protocol_name)
    # to figure out presence of sessions.
    ses_markers: list[str] = []

    # there might be fixups needed so we could deduce session etc
    # this copy is not replacing original one, so the same fix_seqinfo
    # might be called later
    seqinfo_lst = fix_seqinfo(seqinfo_lst)
    for s in seqinfo_lst:
        if s.is_derived:
            continue
        session_ = parse_series_spec(s.protocol_name).get("session", None)
        if session_ and "{" in session_:
            # there was a marker for something we could provide from our seqinfo
            # e.g. {date}
            session_ = session_.format(**s._asdict())
        if session_:
            ses_markers.append(session_)
    session: Optional[str] = None
    if ses_markers:
        # we have a session or possibly more than one even
        # let's figure out which case we have
        nonsign_vals = set(ses_markers).difference("+=")
        # although we might want an explicit '=' to note the same session as
        # mentioned before?
        if len(nonsign_vals) > 1:
            lgr.warning(  # raise NotImplementedError(
                "Cannot deal with multiple sessions in the same study yet!"
                " We will process until the end of the first session"
            )
        if nonsign_vals:
            # get only unique values
            ses_markers = list(set(ses_markers))
            if set(ses_markers).intersection("+="):
                raise NotImplementedError(
                    "Should not mix hardcoded session markers with incremental ones (+=)"
                )
            if not len(ses_markers) == 1:
                raise NotImplementedError(
                    "Should have got a single session marker.  Got following: %s"
                    % ", ".join(map(repr, ses_markers))
                )
            session = ses_markers[0]
        else:
            # TODO - I think we are doomed to go through the sequence and split
            # ... actually the same as with nonsign_vals, we just would need to figure
            # out initial one if sign ones, and should make use of knowing
            # outdir
            # raise NotImplementedError()
            # we need to look at what sessions we already have
            sessions_dir = os.path.join(outdir, locator, "sub-" + subject)
            prior_sessions = sorted(glob(os.path.join(sessions_dir, "ses-*")))
            # TODO: more complicated logic
            # For now just increment session if + and keep the same number if =
            # and otherwise just give it 001
            # Note: this disables our safety blanket which would refuse to process
            # what was already processed before since it would try to override,
            # BUT there is no other way besides only if heudiconv was storing
            # its info based on some UID
            if ses_markers == ["+"]:
                session = "%03d" % (len(prior_sessions) + 1)
            elif ses_markers == ["="]:
                session = (
                    os.path.basename(prior_sessions[-1])[4:]
                    if prior_sessions
                    else "001"
                )
            else:
                session = "001"

    if study_description_hash == "9d148e2a05f782273f6343507733309d":
        session = "siemens1"
        lgr.info("Imposing session {0}".format(session))

    return {
        # TODO: request info on study from the JedCap
        "locator": locator,
        # Sessions to be deduced yet from the names etc TODO
        "session": session,
        "subject": subject,
    }


def sanitize_str(value: str) -> str:
    """Remove illegal characters for BIDS from task/acq/etc.."""
    return _delete_chars(value, "#!@$%^&.,:;_-")


def parse_series_spec(series_spec: str) -> dict[str, str]:
    """Parse protocol name according to our convention with minimal set of fixups"""
    # Since Yarik didn't know better place to put it in, but could migrate outside
    # at some point. TODO
    series_spec = series_spec.replace("anat_T1w", "anat-T1w")
    series_spec = series_spec.replace("hardi_64", "dwi_acq-hardi64")
    series_spec = series_spec.replace("AAHead_Scout", "anat-scout")

    # Parse the name according to our convention/specification

    # leading or trailing spaces do not matter
    series_spec = series_spec.strip(" ")

    # Strip off leading CAPITALS: prefix to accommodate some reported usecases:
    # https://github.com/ReproNim/reproin/issues/14
    # where PU: prefix is added by the scanner
    series_spec = re.sub("^[A-Z]*:", "", series_spec)
    series_spec = re.sub("^WIP ", "", series_spec)  # remove Philips WIP prefix

    # Remove possible suffix we don't care about after __
    series_spec = series_spec.split("__", 1)[0]

    bids = False  # we don't know yet for sure
    # We need to figure out if it is a valid bids
    split = series_spec.split("_")
    prefix = split[0]

    # Fixups
    if prefix == "scout":
        prefix = split[0] = "anat-scout"

    if prefix != "bids" and "-" in prefix:
        prefix, _ = prefix.split("-", 1)
    if prefix == "bids":
        bids = True  # for sure
        split = split[1:]

    def split2(s: str) -> tuple[str, Optional[str]]:
        # split on - if present, if not -- 2nd one returned None
        if "-" in s:
            a, _, b = s.partition("-")
            return a, b
        return s, None

    # Let's analyze first element which should tell us sequence type
    datatype, datatype_suffix = split2(split[0])
    if datatype not in KNOWN_DATATYPES:
        # It is not something we don't consume
        if bids:
            lgr.warning(
                "It was instructed to be BIDS datatype but unknown "
                "%s found. Known are: %s",
                datatype,
                ", ".join(KNOWN_DATATYPES),
            )
        return {}

    regd = dict(datatype=datatype)
    if datatype_suffix:
        regd["datatype_suffix"] = datatype_suffix
    # now go through each to see if one which we care
    bids_leftovers = []
    for s in split[1:]:
        key, value = split2(s)
        if value is None and key[-1] in "+=":
            value = key[-1]
            key = key[:-1]

        # sanitize values, which must not have _ and - is undesirable ATM as well
        # TODO: BIDSv2.0 -- allows "-" so replace with it instead
        value = (
            str(value)
            .replace("_", "X")
            .replace("-", "X")
            .replace("(", "{")
            .replace(")", "}")
        )  # for Philips

        if key in ["ses", "run", "task", "acq", "dir"]:
            # those we care about explicitly
            regd[{"ses": "session"}.get(key, key)] = sanitize_str(value)
        else:
            bids_leftovers.append(s)

    if bids_leftovers:
        regd["bids"] = "_".join(bids_leftovers)

    # TODO: might want to check for all known "standard" BIDS suffixes here
    # among bids_leftovers, thus serve some kind of BIDS validator

    # if not regd.get('datatype_suffix', None):
    #     # might need to assign a default label for each datatype if was not
    #     # given
    #     regd['datatype_suffix'] = {
    #         'func': 'bold'
    #     }.get(regd['datatype'], None)

    return regd


def fixup_subjectid(subjectid: str) -> str:
    """Just in case someone managed to miss a zero or added an extra one"""
    # make it lowercase
    subjectid = subjectid.lower()
    reg = re.match(r"sid0*(\d+)$", subjectid)
    if not reg:
        # some completely other pattern
        # just filter out possible _- in it
        return re.sub("[-_]", "", subjectid)
    return "sid%06d" % int(reg.groups()[0])



from __future__ import annotations

from typing import Optional

from heudiconv.utils import SeqInfo

# Dictionary to specify options for the `populate_intended_for`.
# Valid options are defined in 'bids.py' (for 'matching_parameters':
# ['Shims', 'ImagingVolume',]; for 'criterion': ['First', 'Closest']
POPULATE_INTENDED_FOR_OPTS = {
    "matching_parameters": "ImagingVolume",
    "criterion": "Closest",
}


def create_key(
    template: Optional[str],
    outtype: tuple[str, ...] = ("nii.gz",),
    annotation_classes: None = None,
) -> tuple[str, tuple[str, ...], None]:
    if template is None or not template:
        raise ValueError("Template must be a valid format string")
    return (template, outtype, annotation_classes)


def infotodict(
    seqinfo: list[SeqInfo],
) -> dict[tuple[str, tuple[str, ...], None], list]:
    """Heuristic evaluator for determining which runs belong where

    allowed template fields - follow python string module:

    item: index within category
    subject: participant id
    seqitem: run number during scanning
    subindex: sub index within group
    """

    rs = create_key("rsfmri/rest_run{item:03d}/rest", outtype=("dicom", "nii.gz"))
    boldt1 = create_key("BOLD/task001_run{item:03d}/bold")
    boldt2 = create_key("BOLD/task002_run{item:03d}/bold")
    boldt3 = create_key("BOLD/task003_run{item:03d}/bold")
    boldt4 = create_key("BOLD/task004_run{item:03d}/bold")
    boldt5 = create_key("BOLD/task005_run{item:03d}/bold")
    boldt6 = create_key("BOLD/task006_run{item:03d}/bold")
    boldt7 = create_key("BOLD/task007_run{item:03d}/bold")
    boldt8 = create_key("BOLD/task008_run{item:03d}/bold")
    fm1 = create_key("fieldmap/fm1_{item:03d}")
    fm2 = create_key("fieldmap/fm2_{item:03d}")
    fmrest = create_key("fieldmap/fmrest_{item:03d}")
    dwi = create_key("dmri/dwi_{item:03d}", outtype=("dicom", "nii.gz"))
    t1 = create_key("anatomy/T1_{item:03d}")
    asl = create_key("rsfmri/asl_run{item:03d}/asl")
    aslcal = create_key("rsfmri/asl_run{item:03d}/cal_{subindex:03d}")
    info: dict[tuple[str, tuple[str, ...], None], list] = {
        rs: [],
        boldt1: [],
        boldt2: [],
        boldt3: [],
        boldt4: [],
        boldt5: [],
        boldt6: [],
        boldt7: [],
        boldt8: [],
        fm1: [],
        fm2: [],
        fmrest: [],
        dwi: [],
        t1: [],
        asl: [],
        aslcal: [[]],
    }
    last_run = len(seqinfo)
    for s in seqinfo:
        series_num_str = s.series_id.split("-", 1)[0]
        if not series_num_str.isdecimal():
            raise ValueError(
                f"This heuristic can operate only on data when series_id has form <series-number>-<something else>, "
                f"and <series-number> is a numeric number. Got series_id={s.series_id}"
            )
        series_num: int = int(series_num_str)
        sl, nt = (s.dim3, s.dim4)
        if (sl == 176) and (nt == 1) and ("MPRAGE" in s.protocol_name):
            info[t1] = [s.series_id]
        elif (nt > 60) and ("ge_func_2x2x2_Resting" in s.protocol_name):
            if not s.is_motion_corrected:
                info[rs].append(s.series_id)
        elif (
            (nt == 156)
            and ("ge_functionals_128_PACE_ACPC-30" in s.protocol_name)
            and series_num < last_run
        ):
            if not s.is_motion_corrected:
                info[boldt1].append(s.series_id)
                last_run = series_num
        elif (nt == 155) and ("ge_functionals_128_PACE_ACPC-30" in s.protocol_name):
            if not s.is_motion_corrected:
                info[boldt2].append(s.series_id)
        elif (nt == 222) and ("ge_functionals_128_PACE_ACPC-30" in s.protocol_name):
            if not s.is_motion_corrected:
                info[boldt3].append(s.series_id)
        elif (nt == 114) and ("ge_functionals_128_PACE_ACPC-30" in s.protocol_name):
            if not s.is_motion_corrected:
                info[boldt4].append(s.series_id)
        elif (nt == 156) and ("ge_functionals_128_PACE_ACPC-30" in s.protocol_name):
            if not s.is_motion_corrected and (series_num > last_run):
                info[boldt5].append(s.series_id)
        elif (nt == 324) and ("ge_func_3.1x3.1x4_PACE" in s.protocol_name):
            if not s.is_motion_corrected:
                info[boldt6].append(s.series_id)
        elif (nt == 250) and ("ge_func_3.1x3.1x4_PACE" in s.protocol_name):
            if not s.is_motion_corrected:
                info[boldt7].append(s.series_id)
        elif (nt == 136) and ("ge_func_3.1x3.1x4_PACE" in s.protocol_name):
            if not s.is_motion_corrected:
                info[boldt8].append(s.series_id)
        elif (nt == 101) and ("ep2d_pasl_FairQuipssII" in s.protocol_name):
            if not s.is_motion_corrected:
                info[asl].append(s.series_id)
        elif (nt == 1) and ("ep2d_pasl_FairQuipssII" in s.protocol_name):
            info[aslcal][0].append(s.series_id)
        elif (sl > 1) and (nt == 70) and ("DIFFUSION" in s.protocol_name):
            info[dwi].append(s.series_id)
        elif "field_mapping_128" in s.protocol_name:
            info[fm1].append(s.series_id)
        elif "field_mapping_3.1" in s.protocol_name:
            info[fm2].append(s.series_id)
        elif "field_mapping_Resting" in s.protocol_name:
            info[fmrest].append(s.series_id)
        else:
            pass
    return info



from __future__ import annotations

from glob import glob
import logging
import os.path as op
import sys
from types import TracebackType
from typing import Any, Optional

from . import __packagename__, __version__
from .bids import populate_bids_templates, populate_intended_for, tuneup_bids_json_files
from .convert import prep_conversion
from .due import Doi, due
from .parser import get_study_sessions
from .queue import queue_conversion
from .utils import SeqInfo, anonymize_sid, load_heuristic, treat_infofile

lgr = logging.getLogger(__name__)


INIT_MSG = "Running {packname} version {version} latest {latest}".format


def is_interactive() -> bool:
    """Return True if all in/outs are tty"""
    # TODO: check on windows if hasattr check would work correctly and add value:
    return sys.stdin.isatty() and sys.stdout.isatty() and sys.stderr.isatty()


def setup_exceptionhook() -> None:
    """
    Overloads default sys.excepthook with our exceptionhook handler.

    If interactive, our exceptionhook handler will invoke pdb.post_mortem;
    if not interactive, then invokes default handler.
    """

    def _pdb_excepthook(
        exc_type: type[BaseException],
        exc_value: BaseException,
        tb: Optional[TracebackType],
    ) -> None:
        if is_interactive():
            import pdb
            import traceback

            traceback.print_exception(exc_type, exc_value, tb)
            # print()
            pdb.post_mortem(tb)
        else:
            lgr.warning("We cannot setup exception hook since not in interactive mode")

    sys.excepthook = _pdb_excepthook


def process_extra_commands(
    outdir: str,
    command: str,
    files: Optional[list[str]],
    heuristic: Optional[str],
    session: Optional[str],
    subjs: Optional[list[str]],
    grouping: str,
) -> None:
    """
    Perform custom command instead of regular operations. Supported commands:
    ['treat-json', 'ls', 'populate-templates', 'populate-intended-for']

    Parameters
    ----------
    outdir : str
        Output directory
    command : {'treat-json', 'ls', 'populate-templates', 'populate-intended-for'}
        Heudiconv command to run
    files : list of str or None
        List of files if command needs/expects it
    heuristic : str or None
        Path to heuristic file or name of builtin heuristic.
    session : str or None
        Session identifier
    subjs : None or list of str
        List of subject identifiers
    grouping : {'studyUID', 'accession_number', 'all', 'custom'}
        How to group dicoms.
    """

    def ensure_has_files() -> None:
        if files is None:
            raise ValueError(f"command {command} expects --files being provided")

    if command == "treat-jsons":
        ensure_has_files()
        assert files is not None  # for mypy now
        for fname in files:
            treat_infofile(fname)
    elif command == "ls":
        ensure_heuristic_arg(heuristic)
        assert heuristic is not None
        ensure_has_files()
        assert files is not None  # for mypy now
        heuristic_mod = load_heuristic(heuristic)
        heuristic_ls = getattr(heuristic_mod, "ls", None)
        for fname in files:
            study_sessions = get_study_sessions(
                None,
                [fname],
                heuristic_mod,
                outdir,
                session,
                subjs,
                grouping=grouping,
            )
            print(fname)
            for study_session, sequences in study_sessions.items():
                assert isinstance(sequences, dict)
                suf = ""
                if heuristic_ls:
                    suf += heuristic_ls(study_session, list(sequences.keys()))
                print("\t%s %d sequences%s" % (str(study_session), len(sequences), suf))
    elif command == "populate-templates":
        ensure_heuristic_arg(heuristic)
        assert heuristic is not None
        ensure_has_files()
        assert files is not None  # for mypy now
        heuristic_mod = load_heuristic(heuristic)
        for fname in files:
            populate_bids_templates(fname, getattr(heuristic_mod, "DEFAULT_FIELDS", {}))
    elif command == "sanitize-jsons":
        ensure_has_files()
        assert files is not None  # for mypy now
        tuneup_bids_json_files(files)
    elif command == "heuristics":
        from .utils import get_known_heuristics_with_descriptions

        for name_desc in get_known_heuristics_with_descriptions().items():
            print("- %s: %s" % name_desc)
    elif command == "heuristic-info":
        ensure_heuristic_arg(heuristic)
        assert heuristic is not None
        from .utils import get_heuristic_description

        print(get_heuristic_description(heuristic, full=True))
    elif command == "populate-intended-for":
        kwargs: dict[str, Any] = {}
        if heuristic:
            heuristic_mod = load_heuristic(heuristic)
            kwargs = getattr(heuristic_mod, "POPULATE_INTENDED_FOR_OPTS", {})
        if not subjs:
            subjs = [
                # search outdir for 'sub-*'; if it is a directory (not a regular file), remove
                # the initial 'sub-':
                op.basename(s)[len("sub-") :]
                for s in glob(op.join(outdir, "sub-*"))
                if op.isdir(s)
            ]
            # read the subjects from the participants.tsv file to compare:
            participants_tsv = op.join(outdir, "participants.tsv")
            if op.lexists(participants_tsv):
                with open(participants_tsv, "r") as f:
                    # read header line and find index for 'participant_id':
                    participant_id_index = (
                        f.readline().split("\t").index("participant_id")
                    )
                    # read all participants, removing the initial 'sub-':
                    known_subjects = [
                        ln.split("\t")[participant_id_index][len("sub-") :]
                        for ln in f.readlines()
                    ]
                if not set(subjs) == set(known_subjects):
                    # issue a warning, but continue with the 'subjs' list (the subjects for
                    # which there is data):
                    lgr.warning(
                        "'participants.tsv' contents are not identical to subjects found "
                        "in the BIDS dataset %s",
                        outdir,
                    )

        for subj in subjs:
            subject_path = op.join(outdir, "sub-" + subj)
            if session:
                session_paths = [op.join(subject_path, "ses-" + session)]
            else:
                # check to see if the data for this subject is organized by sessions; if not
                # just use the subject_path
                session_paths = [
                    s for s in glob(op.join(subject_path, "ses-*")) if op.isdir(s)
                ] or [subject_path]
            for session_path in session_paths:
                populate_intended_for(session_path, **kwargs)
    else:
        raise ValueError("Unknown command %s" % command)


def ensure_heuristic_arg(heuristic: Optional[str] = None) -> None:
    """
    Check that the heuristic argument was provided.
    """
    from .utils import get_known_heuristic_names

    if not heuristic:
        raise ValueError(
            "Specify heuristic using -f. Known are: %s"
            % ", ".join(get_known_heuristic_names())
        )


@due.dcite(
    Doi("10.5281/zenodo.1012598"),
    path="heudiconv",
    description="Flexible DICOM converter for organizing brain imaging data",
    version=__version__,
    cite_module=True,
)
def workflow(
    *,
    dicom_dir_template: Optional[str] = None,
    files: Optional[list[str]] = None,
    subjs: Optional[list[str]] = None,
    converter: str = "dcm2niix",
    outdir: str = ".",
    locator: Optional[str] = None,
    conv_outdir: Optional[str] = None,
    anon_cmd: Optional[str] = None,
    heuristic: Optional[str] = None,
    with_prov: bool = False,
    session: Optional[str] = None,
    bids_options: Optional[str] = None,
    overwrite: bool = False,
    datalad: bool = False,
    debug: bool = False,
    command: Optional[str] = None,
    grouping: str = "studyUID",
    minmeta: bool = False,
    random_seed: Optional[int] = None,
    dcmconfig: Optional[str] = None,
    queue: Optional[str] = None,
    queue_args: Optional[str] = None,
) -> None:
    """Run the HeuDiConv conversion workflow.

    Parameters
    ----------
    dicom_dir_template : str or None, optional
        Location of dicomdir that can be indexed with subject id
        {subject} and session {session}. Tarballs (can be compressed)
        are supported in addition to directory. All matching tarballs
        for a subject are extracted and their content processed in a
        single pass. If multiple tarballs are found, each is assumed to
        be a separate session and the 'session' argument is ignored.
        Mutually exclusive with 'files'. Default is None.
    files : list or None, optional
        Files (tarballs, dicoms) or directories containing files to
        process. Mutually exclusive with 'dicom_dir_template'. Default is None.
    subjs : list or None, optional
        List of subjects - required for dicom template. If not
        provided, DICOMS would first be "sorted" and subject IDs
        deduced by the heuristic. Default is None.
    converter : {'dcm2niix', 'none'}, optional
        Tool to use for DICOM conversion. Setting to 'none' disables
        the actual conversion step -- useful for testing heuristics.
        Default is 'dcm2niix'.
    outdir : str, optional
        Output directory for conversion setup (for further
        customization and future reference. This directory will refer
        to non-anonymized subject IDs.
        Default is '.' (current working directory).
    locator : str or 'unknown' or None, optional
        Study path under outdir. If provided, it overloads the value
        provided by the heuristic. If 'datalad=True', every
        directory within locator becomes a super-dataset thus
        establishing a hierarchy. Setting to "unknown" will skip that
        dataset. Default is None.
    conv_outdir : str or None, optional
        Output directory for converted files. By default this is
        identical to --outdir. This option is most useful in
        combination with 'anon_cmd'. Default is None.
    anon_cmd : str or None, optional
        Command to run to convert subject IDs used for DICOMs to
        anonymized IDs. Such command must take a single argument and
        return a single anonymized ID. Also see 'conv_outdir'. Default is None.
    heuristic : str or None, optional
        Name of a known heuristic or path to the Python script containing
        heuristic. Default is None.
    with_prov : bool, optional
        Store additional provenance information. Requires python-rdflib.
        Default is False.
    session : str or None, optional
        Session for longitudinal study_sessions. Default is None.
    bids_options : str or None, optional
        Flag for output into BIDS structure. Can also take BIDS-
        specific options, e.g., --bids notop. The only currently
        supported options is "notop", which skips creation of
        top-level BIDS files. This is useful when running in batch
        mode to prevent possible race conditions. Default is None.
    overwrite : bool, optional
        Overwrite existing converted files. Default is False.
    datalad : bool, optional
        Store the entire collection as DataLad dataset(s). Small files
        will be committed directly to git, while large to annex. New
        version (6) of annex repositories will be used in a "thin"
        mode so it would look to mortals as just any other regular
        directory (i.e. no symlinks to under .git/annex). For now just
        for BIDS mode. Default is False.
    debug : bool, optional
        Do not catch exceptions and show exception traceback. Default is False.
    command : {'heuristics', 'heuristic-info', 'ls', 'populate-templates',
               'sanitize-jsons', 'treat-jsons', 'populate-intended-for', None}, optional
        Custom action to be performed on provided files instead of regular
        operation. Default is None.
    grouping : {'studyUID', 'accession_number', 'all', 'custom'}, optional
        How to group dicoms. Default is 'studyUID'.
    minmeta : bool, optional
        Exclude dcmstack meta information in sidecar jsons. Default is False.
    random_seed : int or None, optional
        Random seed to initialize RNG. Default is None.
    dcmconfig : str or None, optional
        JSON file for additional dcm2niix configuration. Default is None.
    queue : {'SLURM', None}, optional
        Batch system to submit jobs in parallel. Default is None.
        If set, will cause scheduling of conversion and return without performing
        any further action.
    queue_args : str or None, optional
        Additional queue arguments passed as single string of space-separated
        Argument=Value pairs. Default is None.

    Notes
    -----
    All parameters in this function must be called as keyword arguments.
    """

    # To be done asap so anything random is deterministic
    if random_seed is not None:
        import random

        random.seed(random_seed)
        import numpy

        numpy.random.seed(random_seed)
    # Ensure only supported bids options are passed
    if debug:
        lgr.setLevel(logging.DEBUG)
    # Should be possible but only with a single subject -- will be used to
    # override subject deduced from the DICOMs
    if files and subjs and len(subjs) > 1:
        raise ValueError("Unable to processes multiple `--subjects` with files")

    if debug:
        setup_exceptionhook()

    # Deal with provided files or templates
    # pre-process provided list of files and possibly sort into groups/sessions
    # Group files per each study/sid/session

    outdir = op.abspath(outdir)

    latest = None
    try:
        import etelemetry

        latest = etelemetry.get_project("nipy/heudiconv")
    except Exception as e:
        lgr.warning("Could not check for version updates: %s", str(e))

    lgr.info(
        INIT_MSG(
            packname=__packagename__,
            version=__version__,
            latest=(latest or {}).get("version", "Unknown"),
        )
    )

    if command:
        if dicom_dir_template:
            lgr.warning(
                f"DICOM directory template {dicom_dir_template!r} was provided but will be ignored since "
                f"commands do not care about it ATM"
            )

        process_extra_commands(
            outdir,
            command,
            files,
            heuristic,
            session,
            subjs,
            grouping,
        )
        return
    #
    # Load heuristic -- better do it asap to make sure it loads correctly
    #
    if not heuristic:
        raise RuntimeError("No heuristic specified - add to arguments and rerun")

    if queue:
        lgr.info("Queuing %s conversion", queue)
        if files:
            iterarg = "files"
            iterables = len(files)
        elif subjs:
            iterarg = "subjects"
            iterables = len(subjs)
        else:
            raise ValueError("'queue' given but both 'files' and 'subjects' are false")
        queue_conversion(queue, iterarg, iterables, queue_args)
        return

    heuristic_mod = load_heuristic(heuristic)

    study_sessions = get_study_sessions(
        dicom_dir_template,
        files,
        heuristic_mod,
        outdir,
        session,
        subjs,
        grouping=grouping,
    )

    # extract tarballs, and replace their entries with expanded lists of files
    # TODO: we might need to sort so sessions are ordered???
    lgr.info("Need to process %d study sessions", len(study_sessions))

    # processed_studydirs = set()

    locator_manual, session_manual = locator, session
    for (locator, session_, sid), files_or_seqinfo in study_sessions.items():
        # Allow for session to be overloaded from command line
        if session_manual is not None:
            session_ = session_manual
        if locator_manual is not None:
            locator = locator_manual
        if not len(files_or_seqinfo):
            raise ValueError("nothing to process?")
        # that is how life is ATM :-/ since we don't do sorting if subj
        # template is provided
        if isinstance(files_or_seqinfo, dict):
            assert isinstance(list(files_or_seqinfo.keys())[0], SeqInfo)
            dicoms = None
            seqinfo = files_or_seqinfo
        else:
            dicoms = files_or_seqinfo
            seqinfo = None

        if locator == "unknown":
            lgr.warning("Skipping unknown locator dataset")
            continue

        if anon_cmd and sid is not None:
            anon_sid = anonymize_sid(sid, anon_cmd)
            lgr.info("Anonymized {} to {}".format(sid, anon_sid))
        else:
            anon_sid = None

        study_outdir = op.join(outdir, locator or "")
        anon_outdir = conv_outdir or outdir
        anon_study_outdir = op.join(anon_outdir, locator or "")

        if datalad:
            from .external.dlad import prepare_datalad

            dlad_sid = sid if not anon_sid else anon_sid
            dl_msg = prepare_datalad(
                anon_study_outdir,
                anon_outdir,
                dlad_sid,
                session_,
                seqinfo,
                dicoms,
                bids_options,
            )

        lgr.info(
            "PROCESSING STARTS: {0}".format(
                str(dict(subject=sid, outdir=study_outdir, session=session_))
            )
        )

        prep_conversion(
            sid,
            dicoms,
            study_outdir,
            heuristic_mod,
            converter=converter,
            anon_sid=anon_sid,
            anon_outdir=anon_study_outdir,
            with_prov=with_prov,
            ses=session_,
            bids_options=bids_options,
            seqinfo=seqinfo,
            min_meta=minmeta,
            overwrite=overwrite,
            dcmconfig=dcmconfig,
            grouping=grouping,
        )

        lgr.info(
            "PROCESSING DONE: {0}".format(
                str(dict(subject=sid, outdir=study_outdir, session=session_))
            )
        )

        if datalad:
            from .external.dlad import add_to_datalad

            msg = "Converted subject %s" % dl_msg
            # TODO:  whenever propagate to supers work -- do just
            # ds.save(msg=msg)
            #  also in batch mode might fail since we have no locking ATM
            #  and theoretically no need actually to save entire study
            #  we just need that
            add_to_datalad(outdir, study_outdir, msg, bids_options)

    # if bids:
    #     # Let's populate BIDS templates for folks to take care about
    #     for study_outdir in processed_studydirs:
    #         populate_bids_templates(study_outdir)
    #
    # TODO: record_collection of the sid/session although that information
    # is pretty much present in .heudiconv/SUBJECT/info so we could just poke there



from __future__ import annotations

import datetime
from glob import glob
import json
import os.path as op
from pathlib import Path

import pydicom as dcm
import pytest

from heudiconv.cli.run import main as runner
from heudiconv.convert import nipype_convert
from heudiconv.dicoms import (
    embed_dicom_and_nifti_metadata,
    get_datetime_from_dcm,
    get_reproducible_int,
    group_dicoms_into_seqinfos,
    parse_private_csa_header,
)

from .utils import TESTS_DATA_PATH

# Public: Private DICOM tags
DICOM_FIELDS_TO_TEST = {"ProtocolName": "tProtocolName"}


def test_private_csa_header(tmp_path: Path) -> None:
    dcm_file = op.join(TESTS_DATA_PATH, "axasc35.dcm")
    dcm_data = dcm.dcmread(dcm_file, stop_before_pixels=True)
    for pub, priv in DICOM_FIELDS_TO_TEST.items():
        # ensure missing public tag
        with pytest.raises(AttributeError):
            getattr(dcm, pub)
        # ensure private tag is found
        assert parse_private_csa_header(dcm_data, pub, priv) != ""
        # and quickly run heudiconv with no conversion
        runner(
            ["--files", dcm_file, "-c", "none", "-f", "reproin", "-o", str(tmp_path)]
        )


def test_embed_dicom_and_nifti_metadata(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    """Test dcmstack's additional fields"""
    monkeypatch.chdir(tmp_path)
    # set up testing files
    dcmfiles = [op.join(TESTS_DATA_PATH, "axasc35.dcm")]
    infofile = "infofile.json"

    out_prefix = str(tmp_path / "nifti")
    # 1) nifti does not exist -- no longer supported
    with pytest.raises(NotImplementedError):
        embed_dicom_and_nifti_metadata(dcmfiles, out_prefix + ".nii.gz", infofile, None)

    # we should produce nifti using our "standard" ways
    nipype_out, prov_file = nipype_convert(
        dcmfiles,
        prefix=out_prefix,
        with_prov=False,
        bids_options=None,
        tmpdir=str(tmp_path),
    )
    niftifile = nipype_out.outputs.converted_files

    assert op.exists(niftifile)
    # 2) nifti exists
    embed_dicom_and_nifti_metadata(dcmfiles, niftifile, infofile, None)
    assert op.exists(infofile)
    with open(infofile) as fp:
        out2 = json.load(fp)

    # 3) with existing metadata
    bids = {"existing": "data"}
    embed_dicom_and_nifti_metadata(dcmfiles, niftifile, infofile, bids)
    with open(infofile) as fp:
        out3 = json.load(fp)

    assert out3.pop("existing") == "data"
    assert out3 == out2


def test_group_dicoms_into_seqinfos() -> None:
    """Tests for group_dicoms_into_seqinfos"""

    # 1) Check that it works for PhoenixDocuments:
    # set up testing files
    dcmfolder = op.join(TESTS_DATA_PATH, "Phoenix")
    dcmfiles = glob(op.join(dcmfolder, "*", "*.dcm"))

    seqinfo = group_dicoms_into_seqinfos(dcmfiles, "studyUID", flatten=True)

    assert type(seqinfo) is dict
    assert len(seqinfo) == len(dcmfiles)
    assert [s.series_description for s in seqinfo] == [
        "AAHead_Scout_32ch-head-coil",
        "PhoenixZIPReport",
    ]


def test_custom_seqinfo() -> None:
    """Tests for custom seqinfo extraction"""

    from heudiconv.heuristics.convertall_custom import custom_seqinfo

    dcmfiles = glob(op.join(TESTS_DATA_PATH, "phantom.dcm"))

    seqinfos = group_dicoms_into_seqinfos(
        dcmfiles, "studyUID", flatten=True, custom_seqinfo=custom_seqinfo
    )  # type: ignore

    seqinfo = list(seqinfos.keys())[0]

    assert hasattr(seqinfo, "custom")
    assert isinstance(seqinfo.custom, tuple)
    assert len(seqinfo.custom) == 2
    assert seqinfo.custom[1] == dcmfiles[0]


def test_get_datetime_from_dcm_from_acq_date_time() -> None:
    typical_dcm = dcm.dcmread(
        op.join(TESTS_DATA_PATH, "phantom.dcm"), stop_before_pixels=True
    )

    # do we try to grab from AcquisitionDate/AcquisitionTime first when available?
    dt = get_datetime_from_dcm(typical_dcm)
    assert dt == datetime.datetime.strptime(
        typical_dcm.get("AcquisitionDate") + typical_dcm.get("AcquisitionTime"),
        "%Y%m%d%H%M%S.%f",
    )


def test_get_datetime_from_dcm_from_acq_datetime() -> None:
    # if AcquisitionDate and AcquisitionTime not there, can we rely on AcquisitionDateTime?
    XA30_enhanced_dcm = dcm.dcmread(
        op.join(
            TESTS_DATA_PATH,
            "MRI_102TD_PHA_S.MR.Chen_Matthews_1.3.1.2022.11.16.15.50.20.357.31204541.dcm",
        ),
        stop_before_pixels=True,
    )
    dt = get_datetime_from_dcm(XA30_enhanced_dcm)
    assert dt == datetime.datetime.strptime(
        XA30_enhanced_dcm.get("AcquisitionDateTime"), "%Y%m%d%H%M%S.%f"
    )


def test_get_datetime_from_dcm_from_only_series_date_time() -> None:
    # if acquisition date/time/datetime not available, can we rely on SeriesDate & SeriesTime?
    XA30_enhanced_dcm = dcm.dcmread(
        op.join(
            TESTS_DATA_PATH,
            "MRI_102TD_PHA_S.MR.Chen_Matthews_1.3.1.2022.11.16.15.50.20.357.31204541.dcm",
        ),
        stop_before_pixels=True,
    )
    del XA30_enhanced_dcm.AcquisitionDateTime
    dt = get_datetime_from_dcm(XA30_enhanced_dcm)
    assert dt == datetime.datetime.strptime(
        XA30_enhanced_dcm.get("SeriesDate") + XA30_enhanced_dcm.get("SeriesTime"),
        "%Y%m%d%H%M%S.%f",
    )


def test_get_datetime_from_dcm_wo_dt() -> None:
    # if there's no known source (e.g., after anonymization), are we still able to proceed?
    XA30_enhanced_dcm = dcm.dcmread(
        op.join(
            TESTS_DATA_PATH,
            "MRI_102TD_PHA_S.MR.Chen_Matthews_1.3.1.2022.11.16.15.50.20.357.31204541.dcm",
        ),
        stop_before_pixels=True,
    )
    del XA30_enhanced_dcm.AcquisitionDateTime
    del XA30_enhanced_dcm.SeriesDate
    del XA30_enhanced_dcm.SeriesTime
    assert get_datetime_from_dcm(XA30_enhanced_dcm) is None


def test_get_reproducible_int() -> None:
    dcmfile = op.join(TESTS_DATA_PATH, "phantom.dcm")

    assert type(get_reproducible_int([dcmfile])) is int


@pytest.mark.skip(
    reason="This test was mistakenly marked as a fixture, and removing the fixture decorator led to the test failing.  Don't know how to fix."
)
def test_get_reproducible_int_wo_dt(tmp_path: Path) -> None:
    # can this function return an int when we don't have any usable dates?
    typical_dcm = dcm.dcmread(
        op.join(TESTS_DATA_PATH, "phantom.dcm"), stop_before_pixels=True
    )
    del typical_dcm.SeriesDate
    del typical_dcm.AcquisitionDate
    dcm.dcmwrite(tmp_path, typical_dcm)

    assert type(get_reproducible_int([str(tmp_path)])) is int


@pytest.mark.skip(
    reason="This test was mistakenly marked as a fixture, and removing the fixture decorator led to the test failing.  Don't know how to fix."
)
def test_get_reproducible_int_raises_assertion_wo_dt(tmp_path: Path) -> None:
    # if there's no known source (e.g., after anonymization), is AssertionError Raised?
    XA30_enhanced_dcm = dcm.dcmread(
        op.join(
            TESTS_DATA_PATH,
            "MRI_102TD_PHA_S.MR.Chen_Matthews_1.3.1.2022.11.16.15.50.20.357.31204541.dcm",
        ),
        stop_before_pixels=True,
    )
    del XA30_enhanced_dcm.AcquisitionDateTime
    del XA30_enhanced_dcm.SeriesDate
    del XA30_enhanced_dcm.SeriesTime
    dcm.dcmwrite(tmp_path, dataset=XA30_enhanced_dcm)
    with pytest.raises(AssertionError):
        get_reproducible_int([str(tmp_path)])



"""Heuristic demonstrating conversion of the PhoenixZIPReport from Siemens.

It only cares about converting a series with have PhoenixZIPReport in their
series_description and outputs **only to sourcedata**.
"""

from __future__ import annotations

from typing import Optional

from heudiconv.utils import SeqInfo


def create_key(
    template: Optional[str],
    outtype: tuple[str, ...] = ("nii.gz",),
    annotation_classes: None = None,
) -> tuple[str, tuple[str, ...], None]:
    if template is None or not template:
        raise ValueError("Template must be a valid format string")
    return (template, outtype, annotation_classes)


def infotodict(
    seqinfo: list[SeqInfo],
) -> dict[tuple[str, tuple[str, ...], None], list[dict[str, str]]]:
    """Heuristic evaluator for determining which runs belong where

    allowed template fields - follow python string module:

    item: index within category
    subject: participant id
    seqitem: run number during scanning
    subindex: sub index within group
    """
    sbref = create_key(
        "sub-{subject}/func/sub-{subject}_task-QA_sbref",
        outtype=(
            "nii.gz",
            "dicom",
        ),
    )
    scout = create_key(
        "sub-{subject}/anat/sub-{subject}_T1w",
        outtype=(
            "nii.gz",
            "dicom",
        ),
    )
    phoenix_doc = create_key(
        "sub-{subject}/misc/sub-{subject}_phoenix", outtype=("dicom",)
    )

    info: dict[tuple[str, tuple[str, ...], None], list[dict[str, str]]] = {
        sbref: [],
        scout: [],
        phoenix_doc: [],
    }
    for s in seqinfo:
        if (
            "PhoenixZIPReport" in s.series_description
            and s.image_type[3] == "CSA REPORT"
        ):
            info[phoenix_doc].append({"item": s.series_id})
        if "scout" in s.series_description.lower():
            info[scout].append({"item": s.series_id})

    return info



from __future__ import annotations

import logging
import os.path as op
from pathlib import Path
from typing import Optional

import heudiconv.heuristics

HEURISTICS_PATH = op.join(heudiconv.heuristics.__path__[0])
TESTS_DATA_PATH = op.join(op.dirname(__file__), "data")

lgr = logging.getLogger(__name__)


def gen_heudiconv_args(
    datadir: str,
    outdir: str,
    subject: str,
    heuristic_file: str,
    anon_cmd: Optional[str] = None,
    template: Optional[str] = None,
    xargs: Optional[list[str]] = None,
) -> list[str]:
    heuristic = op.realpath(op.join(HEURISTICS_PATH, heuristic_file))

    if template:
        # use --dicom_dir_template
        args = ["-d", op.join(datadir, template)]
    else:
        args = ["--files", datadir]

    args.extend(
        [
            "-c",
            "dcm2niix",
            "-o",
            outdir,
            "-s",
            subject,
            "-f",
            heuristic,
            "--bids",
            "--minmeta",
        ]
    )
    if anon_cmd:
        args += ["--anon-cmd", op.join(op.dirname(__file__), anon_cmd), "-a", outdir]
    if xargs:
        args += xargs

    return args


def fetch_data(tmpdir: str | Path, dataset: str, getpath: Optional[str] = None) -> str:
    """
    Utility function to interface with datalad database.
    Performs datalad `install` and datalad `get` operations.

    Parameters
    ----------
    tmpdir : str or Path
        directory to temporarily store data
    dataset : str
        dataset path from `http://datasets-tests.datalad.org`
    getpath : str [optional]
        exclusive path to get

    Returns
    -------
    targetdir : str
        directory with installed dataset
    """
    from datalad import api

    targetdir = op.join(tmpdir, op.basename(dataset))
    ds = api.install(
        path=targetdir, source="http://datasets-tests.datalad.org/{}".format(dataset)
    )

    getdir = targetdir + (op.sep + getpath if getpath is not None else "")
    ds.get(getdir)
    return targetdir



from __future__ import annotations

from glob import glob
import inspect
import logging
import os
import os.path as op
from typing import TYPE_CHECKING, Optional

from ..info import MIN_DATALAD_VERSION as MIN_VERSION
from ..utils import SeqInfo, create_file_if_missing

lgr = logging.getLogger(__name__)

if TYPE_CHECKING:
    from datalad.api import Dataset


def prepare_datalad(
    studydir: str,
    outdir: str,
    sid: Optional[str],
    session: str | int | None,
    seqinfo: Optional[dict[SeqInfo, list[str]]],
    dicoms: Optional[list[str]],
    bids: Optional[str],
) -> str:
    """Prepare data for datalad"""
    from datalad.api import Dataset

    datalad_msg_suf = " %s" % sid
    if session:
        datalad_msg_suf += ", session %s" % session
    if seqinfo:
        datalad_msg_suf += ", %d sequences" % len(seqinfo)
        datalad_msg_suf += ", %d dicoms" % len(sum(seqinfo.values(), []))
    else:
        assert dicoms is not None
        datalad_msg_suf += ", %d dicoms" % len(dicoms)
    ds = Dataset(studydir)
    if not op.exists(outdir) or not ds.is_installed():
        add_to_datalad(
            outdir, studydir, msg="Preparing for %s" % datalad_msg_suf, bids=bids
        )
    return datalad_msg_suf


def add_to_datalad(
    topdir: str, studydir: str, msg: Optional[str], bids: Optional[str]  # noqa: U100
) -> None:
    """Do all necessary preparations (if were not done before) and save"""
    import datalad.api as dl
    from datalad.api import Dataset
    from datalad.support.annexrepo import AnnexRepo
    from datalad.support.external_versions import external_versions

    assert external_versions["datalad"] >= MIN_VERSION, "Need datalad >= {}".format(
        MIN_VERSION
    )  # add to reqs

    studyrelpath = op.relpath(studydir, topdir)
    assert not studyrelpath.startswith(op.pardir)  # so we are under
    # now we need to test and initiate a DataLad dataset all along the path
    curdir_ = topdir
    superds = None
    subdirs = [""] + [d for d in studyrelpath.split(op.sep) if d != os.curdir]
    for isubdir, subdir in enumerate(subdirs):
        curdir_ = op.join(curdir_, subdir)
        ds = Dataset(curdir_)
        if not ds.is_installed():
            lgr.info("Initiating %s", ds)
            # would require annex > 20161018 for correct operation on annex v6
            # need to add .gitattributes first anyways
            ds_ = dl.create(
                curdir_,
                dataset=superds,
                force=True,
                # initiate annex only at the bottom repository
                annex=isubdir == (len(subdirs) - 1),
                fake_dates=True,
                # shared_access='all',
            )
            assert ds == ds_
        assert ds.is_installed()
        superds = ds

    # TODO: we need a helper (in DataLad ideally) to ease adding such
    # specifications
    gitattributes_path = op.join(studydir, ".gitattributes")
    # We will just make sure that all our desired rules are present in it
    desired_attrs = """\
* annex.largefiles=(largerthan=100kb)
*.json annex.largefiles=nothing
*.txt annex.largefiles=nothing
*.tsv annex.largefiles=nothing
*.nii.gz annex.largefiles=anything
*.tgz annex.largefiles=anything
*_scans.tsv annex.largefiles=anything
"""
    if op.exists(gitattributes_path):
        with open(gitattributes_path, "rb") as f:
            known_attrs = [line.decode("utf-8").rstrip() for line in f.readlines()]
    else:
        known_attrs = []
    for attr in desired_attrs.split("\n"):
        if attr not in known_attrs:
            known_attrs.append(attr)
    with open(gitattributes_path, "wb") as f:
        f.write("\n".join(known_attrs).encode("utf-8"))

    # ds might have memories of having ds.repo GitRepo
    superds = Dataset(topdir)
    assert op.realpath(ds.path) == op.realpath(studydir)
    assert isinstance(ds.repo, AnnexRepo)
    # Add doesn't have all the options of save such as msg and supers
    ds.save(path=[".gitattributes"], message="Custom .gitattributes", to_git=True)
    dsh = dsh_path = None
    if op.lexists(op.join(ds.path, ".heudiconv")):
        dsh_path = op.join(ds.path, ".heudiconv")
        dsh = Dataset(dsh_path)
        if not dsh.is_installed():
            # Previously we did not have it as a submodule, and since no
            # automagic migration is implemented, we just need to check first
            # if any path under .heudiconv is already under git control
            if any(x.startswith(".heudiconv/") for x in ds.repo.get_files()):
                lgr.warning(
                    "%s has .heudiconv not as a submodule from previous"
                    " versions of heudiconv. No automagic migration is "
                    "yet provided",
                    ds,
                )
            else:
                dsh = ds.create(
                    path=".heudiconv",
                    force=True,
                    # shared_access='all'
                )
        # Since .heudiconv could contain sensitive information
        # we place all files under annex and then add
        if create_file_if_missing(
            op.join(dsh_path, ".gitattributes"), """* annex.largefiles=anything"""
        ):
            ds.save(
                ".heudiconv/.gitattributes",
                to_git=True,
                message="Added gitattributes to place all .heudiconv content"
                " under annex",
            )
    save_res = ds.save(
        ".",
        recursive=True
        # not in effect! ?
        # annex_add_opts=['--include-dotfiles']
    )
    annexed_files = [sr["path"] for sr in save_res if sr.get("key", None)]

    # Provide metadata for sensitive information
    sensitive_patterns = [
        "sourcedata",
        "*_scans.tsv",  # top level
        "*/*_scans.tsv",  # within subj
        "*/*/*_scans.tsv",  # within sess/subj
        "*/anat",  # within subj
        "*/*/anat",  # within ses/subj
    ]
    for sp in sensitive_patterns:
        mark_sensitive(ds, sp, annexed_files)
    if dsh_path:
        mark_sensitive(ds, ".heudiconv")  # entire .heudiconv!
    superds.save(path=ds.path, message=msg, recursive=True)

    assert not ds.repo.dirty
    # TODO:  they are still appearing as native annex symlinked beasts
    """
    TODOs:
    it needs
    - unlock  (thin will be in effect)
    - save/commit (does modechange 120000 => 100644
    - could potentially somehow automate that all:
      http://git-annex.branchable.com/tips/automatically_adding_metadata/
    - possibly even make separate sub-datasets for originaldata, derivatives ?
    """


def mark_sensitive(ds: Dataset, path_glob: str, files: list[str] | None = None) -> None:
    """

    Parameters
    ----------
    ds : Dataset to operate on
    path_glob : str
      glob of the paths within dataset to work on
    files : list[str]
      subset of files to mark

    Returns
    -------
    None
    """
    paths = glob(op.join(ds.path, path_glob))
    if files:
        paths = [p for p in paths if p in files]
    if not paths:
        return
    lgr.debug("Marking %d files with distribution-restrictions field", len(paths))
    # set_metadata can be a bloody generator
    res = ds.repo.set_metadata(
        paths, add=dict([("distribution-restrictions", "sensitive")]), recursive=True
    )
    if inspect.isgenerator(res):
        res = list(res)



from __future__ import annotations

import csv
from glob import glob
from io import StringIO
import logging
import os
import os.path as op
from os.path import dirname
from os.path import join as pjoin
from pathlib import Path
import re
from unittest.mock import patch

import pytest

from heudiconv.cli.run import main as runner

from .utils import TESTS_DATA_PATH
from .. import __version__
from ..bids import HEUDICONV_VERSION_JSON_KEY
from ..utils import load_json

lgr = logging.getLogger(__name__)

try:
    from datalad.api import Dataset
except ImportError:  # pragma: no cover
    Dataset = None


# this will fail if not in project's root directory
def test_smoke_convertall(tmp_path: Path) -> None:
    args = [
        "-c",
        "dcm2niix",
        "-o",
        str(tmp_path),
        "-b",
        "--datalad",
        "-s",
        "fmap_acq-3mm",
        "-d",
        f"{TESTS_DATA_PATH}/{{subject}}/*",
    ]

    # complain if no heurisitic
    with pytest.raises(RuntimeError):
        runner(args)

    args.extend(["-f", "convertall"])
    runner(args)


@pytest.mark.parametrize("heuristic", ["reproin", "convertall"])
@pytest.mark.parametrize(
    "invocation",
    [
        ["--files", TESTS_DATA_PATH],  # our new way with automated grouping
        ["-d", f"{TESTS_DATA_PATH}/{{subject}}/*", "-s", "01-fmap_acq-3mm"],
        # "old" way specifying subject
        # should produce the same results
    ],
)
@pytest.mark.skipif(Dataset is None, reason="no datalad")
def test_reproin_largely_smoke(tmp_path: Path, heuristic: str, invocation: str) -> None:
    is_bids = True if heuristic == "reproin" else False
    args = [
        "--random-seed",
        "1",
        "-f",
        heuristic,
        "-c",
        "dcm2niix",
        "-o",
        str(tmp_path),
    ]
    if is_bids:
        args.append("-b")
    args.append("--datalad")
    args.extend(invocation)

    # Test some safeguards
    if invocation[0] == "--files":
        # Multiple subjects must not be specified -- only a single one could
        # be overridden from the command line
        with pytest.raises(ValueError):
            runner(args + ["--subjects", "sub1", "sub2"])

        if heuristic != "reproin":
            # if subject is not overridden, raise error
            with pytest.raises(NotImplementedError):
                runner(args)
            return

    runner(args)
    ds = Dataset(tmp_path)
    assert ds.is_installed()
    assert not ds.repo.dirty
    head = ds.repo.get_hexsha()

    # and if we rerun -- should fail
    lgr.info(
        "RERUNNING, expecting to FAIL since the same everything "
        "and -c specified so we did conversion already"
    )
    with pytest.raises(RuntimeError):
        runner(args)

    # but there should be nothing new
    assert not ds.repo.dirty
    # TODO: remove whenever https://github.com/datalad/datalad/issues/6843
    # is fixed/released
    buggy_datalad = (ds.pathobj / ".gitmodules").read_text().splitlines().count(
        '[submodule "Halchenko"]'
    ) > 1
    assert head == ds.repo.get_hexsha() or buggy_datalad

    # unless we pass 'overwrite' flag
    runner(args + ["--overwrite"])
    # but result should be exactly the same, so it still should be clean
    # and at the same commit
    assert ds.is_installed()
    assert not ds.repo.dirty
    assert head == ds.repo.get_hexsha() or buggy_datalad


@pytest.mark.parametrize(
    "invocation",
    [
        ["--files", TESTS_DATA_PATH],  # our new way with automated grouping
    ],
)
def test_scans_keys_reproin(tmp_path: Path, invocation: list[str]) -> None:
    args = ["-f", "reproin", "-c", "dcm2niix", "-o", str(tmp_path), "-b"]
    args += invocation
    runner(args)
    # for now check it exists
    scans_keys = glob(pjoin(tmp_path, "*/*/*/*/*/*.tsv"))
    assert len(scans_keys) == 1
    with open(scans_keys[0]) as f:
        reader = csv.reader(f, delimiter="\t")
        for i, row in enumerate(reader):
            if i == 0:
                assert row == ["filename", "acq_time", "operator", "randstr"]
            assert len(row) == 4
            if i != 0:
                assert os.path.exists(pjoin(dirname(scans_keys[0]), row[0]))
                assert re.match(
                    r"^[\d]{4}-[\d]{2}-[\d]{2}T[\d]{2}:[\d]{2}:[\d]{2}.[\d]{6}$", row[1]
                )


@patch("sys.stdout", new_callable=StringIO)
def test_ls(stdout: StringIO) -> None:
    args = ["-f", "reproin", "--command", "ls", "--files", TESTS_DATA_PATH]
    runner(args)
    out = stdout.getvalue()
    assert "StudySessionInfo(locator=" in out
    assert "Halchenko/Yarik/950_bids_test4" in out


def test_scout_conversion(tmp_path: Path) -> None:
    args = ["-b", "-f", "reproin", "--files", TESTS_DATA_PATH, "-o", str(tmp_path)]
    runner(args)

    dspath = tmp_path / "Halchenko/Yarik/950_bids_test4"
    sespath = dspath / "sub-phantom1sid1/ses-localizer"

    assert not (sespath / "anat").exists()
    assert (
        dspath / "sourcedata/sub-phantom1sid1/ses-localizer/"
        "anat/sub-phantom1sid1_ses-localizer_scout.dicom.tgz"
    ).exists()

    # Let's do some basic checks on produced files
    j = load_json(
        sespath / "fmap/sub-phantom1sid1_ses-localizer_acq-3mm_phasediff.json"
    )
    # We store HeuDiConv version in each produced .json file
    # TODO: test that we are not somehow overwriting that version in existing
    # files which we have not produced in a particular run.
    assert j[HEUDICONV_VERSION_JSON_KEY] == __version__


@pytest.mark.parametrize(
    "bidsoptions",
    [
        ["notop"],
        [],
    ],
)
def test_notop(tmp_path: Path, bidsoptions: list[str]) -> None:
    args = [
        "-f",
        "reproin",
        "--files",
        TESTS_DATA_PATH,
        "-o",
        str(tmp_path),
        "-b",
    ] + bidsoptions
    runner(args)

    assert op.exists(pjoin(tmp_path, "Halchenko/Yarik/950_bids_test4"))
    for fname in [
        "CHANGES",
        "dataset_description.json",
        "participants.tsv",
        "README",
        "participants.json",
    ]:
        if "notop" in bidsoptions:
            assert not op.exists(
                pjoin(tmp_path, "Halchenko/Yarik/950_bids_test4", fname)
            )
        else:
            assert op.exists(pjoin(tmp_path, "Halchenko/Yarik/950_bids_test4", fname))


def test_phoenix_doc_conversion(tmp_path: Path) -> None:
    subID = "Phoenix"
    args = [
        "-c",
        "dcm2niix",
        "-o",
        str(tmp_path),
        "-b",
        "-f",
        "bids_PhoenixReport",
        "--files",
        pjoin(TESTS_DATA_PATH, "Phoenix"),
        "-s",
        subID,
    ]
    runner(args)

    # check that the Phoenix document has been extracted (as gzipped dicom) in
    # the sourcedata/misc folder:
    assert op.exists(
        pjoin(tmp_path, "sourcedata", "sub-%s", "misc", "sub-%s_phoenix.dicom.tgz")
        % (subID, subID)
    )
    # check that no "sub-<subID>/misc" folder has been created in the BIDS
    # structure:
    assert not op.exists(pjoin(tmp_path, "sub-%s", "misc") % subID)



import logging

from ._version import __version__
from .info import __packagename__

__all__ = ["__packagename__", "__version__"]

lgr = logging.getLogger(__name__)
lgr.debug("Starting the abomination")  # just to "run-test" logging



# dicom operations
from __future__ import annotations

from collections.abc import Callable
import datetime
import logging
import os
import os.path as op
from pathlib import Path
import sys
import tarfile
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Hashable,
    List,
    NamedTuple,
    Optional,
    Protocol,
    Union,
    overload,
)
from unittest.mock import patch
import warnings

import pydicom as dcm

from .utils import (
    SeqInfo,
    TempDirs,
    get_typed_attr,
    load_json,
    set_readonly,
    strptime_micr,
)

if TYPE_CHECKING:
    if sys.version_info >= (3, 8):
        from typing import Literal
    else:
        from typing_extensions import Literal

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    # suppress warning
    import nibabel.nicom.dicomwrappers as dw

lgr = logging.getLogger(__name__)
total_files = 0
# Might be monkey patched by user heuristic to tune desired compression level.
# Preferably do not move/rename.
compresslevel = 9


class CustomSeqinfoT(Protocol):
    def __call__(self, wrapper: dw.Wrapper, series_files: list[str]) -> Hashable:
        ...


def create_seqinfo(
    mw: dw.Wrapper,
    series_files: list[str],
    series_id: str,
    custom_seqinfo: CustomSeqinfoT | None = None,
) -> SeqInfo:
    """Generate sequence info

    Parameters
    ----------
    mw: Wrapper
    series_files: list
    series_id: str
    """
    dcminfo = mw.dcm_data
    accession_number = dcminfo.get("AccessionNumber")

    # TODO: do not group echoes by default
    size: list[int] = list(mw.image_shape) + [len(series_files)]
    if len(size) < 4:
        size.append(1)

    # parse DICOM for seqinfo fields
    TR = get_typed_attr(dcminfo, "RepetitionTime", float, -1000) / 1000
    TE = get_typed_attr(dcminfo, "EchoTime", float, -1)
    refphys = get_typed_attr(dcminfo, "ReferringPhysicianName", str, "")
    image_type = get_typed_attr(dcminfo, "ImageType", tuple, ())
    is_moco = "MOCO" in image_type
    series_desc = get_typed_attr(dcminfo, "SeriesDescription", str, "")

    if dcminfo.get([0x18, 0x24]):
        # GE and Philips
        sequence_name = dcminfo[0x18, 0x24].value
    elif dcminfo.get([0x19, 0x109C]):
        # Siemens
        sequence_name = dcminfo[0x19, 0x109C].value
    else:
        sequence_name = ""

    # initialized in `group_dicoms_to_seqinfos`
    global total_files
    total_files += len(series_files)

    custom_seqinfo_data = (
        custom_seqinfo(wrapper=mw, series_files=series_files)
        if custom_seqinfo
        else None
    )
    try:
        hash(custom_seqinfo_data)
    except TypeError:
        raise RuntimeError(
            "Data returned by the heuristics custom_seqinfo is not hashable. "
            "See https://heudiconv.readthedocs.io/en/latest/heuristics.html#custom_seqinfo for more "
            "details."
        )

    return SeqInfo(
        total_files_till_now=total_files,
        example_dcm_file=op.basename(series_files[0]),
        series_id=series_id,
        dcm_dir_name=op.basename(op.dirname(series_files[0])),
        series_files=len(series_files),
        unspecified="",
        dim1=size[0],
        dim2=size[1],
        dim3=size[2],
        dim4=size[3],
        TR=TR,
        TE=TE,
        protocol_name=dcminfo.ProtocolName,
        is_motion_corrected=is_moco,
        is_derived="derived" in [x.lower() for x in image_type],
        patient_id=dcminfo.get("PatientID"),
        study_description=dcminfo.get("StudyDescription"),
        referring_physician_name=refphys,
        series_description=series_desc,
        sequence_name=sequence_name,
        image_type=image_type,
        accession_number=accession_number,
        # For demographics to populate BIDS participants.tsv
        patient_age=dcminfo.get("PatientAge"),
        patient_sex=dcminfo.get("PatientSex"),
        date=dcminfo.get("AcquisitionDate"),
        series_uid=dcminfo.get("SeriesInstanceUID"),
        time=dcminfo.get("AcquisitionTime"),
        custom=custom_seqinfo_data,
    )


def validate_dicom(
    fl: str, dcmfilter: Optional[Callable[[dcm.dataset.Dataset], Any]]
) -> Optional[tuple[dw.Wrapper, tuple[int, str], Optional[str]]]:
    """
    Parse DICOM attributes. Returns None if not valid.
    """
    mw = dw.wrapper_from_file(fl, force=True, stop_before_pixels=True)
    # clean series signature
    for sig in ("iop", "ICE_Dims", "SequenceName"):
        try:
            del mw.series_signature[sig]
        except KeyError:
            pass
    # Workaround for protocol name in private siemens csa header
    if not getattr(mw.dcm_data, "ProtocolName", "").strip():
        mw.dcm_data.ProtocolName = (
            parse_private_csa_header(mw.dcm_data, "ProtocolName", "tProtocolName")
            if mw.is_csa
            else ""
        )
    try:
        protocol_name = mw.dcm_data.ProtocolName
        assert isinstance(protocol_name, str)
        series_id = (int(mw.dcm_data.SeriesNumber), protocol_name)
    except AttributeError as e:
        lgr.warning('Ignoring %s since not quite a "normal" DICOM: %s', fl, e)
        return None
    if dcmfilter is not None and dcmfilter(mw.dcm_data):
        lgr.warning("Ignoring %s because of DICOM filter", fl)
        return None
    if mw.dcm_data[0x0008, 0x0016].repval in (
        "Raw Data Storage",
        "GrayscaleSoftcopyPresentationStateStorage",
    ):
        return None
    try:
        file_studyUID = mw.dcm_data.StudyInstanceUID
        assert isinstance(file_studyUID, str)
    except AttributeError:
        lgr.info("File {} is missing any StudyInstanceUID".format(fl))
        file_studyUID = None
    return mw, series_id, file_studyUID


class SeriesID(NamedTuple):
    series_number: int
    protocol_name: str
    file_studyUID: Optional[str] = None

    def __str__(self) -> str:
        s = f"{self.series_number}-{self.protocol_name}"
        if self.file_studyUID is not None:
            s += f"-{self.file_studyUID}"
        return s


@overload
def group_dicoms_into_seqinfos(
    files: list[str],
    grouping: str,
    file_filter: Optional[Callable[[str], Any]] = None,
    dcmfilter: Optional[Callable[[dcm.dataset.Dataset], Any]] = None,
    flatten: Literal[False] = False,
    custom_grouping: str
    | Callable[
        [list[str], Optional[Callable[[dcm.dataset.Dataset], Any]], type[SeqInfo]],
        dict[SeqInfo, list[str]],
    ]
    | None = None,
    custom_seqinfo: CustomSeqinfoT | None = None,
) -> dict[Optional[str], dict[SeqInfo, list[str]]]:
    ...


@overload
def group_dicoms_into_seqinfos(
    files: list[str],
    grouping: str,
    file_filter: Optional[Callable[[str], Any]] = None,
    dcmfilter: Optional[Callable[[dcm.dataset.Dataset], Any]] = None,
    *,
    flatten: Literal[True],
    custom_grouping: str
    | Callable[
        [list[str], Optional[Callable[[dcm.dataset.Dataset], Any]], type[SeqInfo]],
        dict[SeqInfo, list[str]],
    ]
    | None = None,
    custom_seqinfo: CustomSeqinfoT | None = None,
) -> dict[SeqInfo, list[str]]:
    ...


def group_dicoms_into_seqinfos(
    files: list[str],
    grouping: str,
    file_filter: Optional[Callable[[str], Any]] = None,
    dcmfilter: Optional[Callable[[dcm.dataset.Dataset], Any]] = None,
    flatten: Literal[False, True] = False,
    custom_grouping: str
    | Callable[
        [list[str], Optional[Callable[[dcm.dataset.Dataset], Any]], type[SeqInfo]],
        dict[SeqInfo, list[str]],
    ]
    | None = None,
    custom_seqinfo: CustomSeqinfoT | None = None,
) -> dict[Optional[str], dict[SeqInfo, list[str]]] | dict[SeqInfo, list[str]]:
    """Process list of dicoms and return seqinfo and file group
    `seqinfo` contains per-sequence extract of fields from DICOMs which
    will be later provided into heuristics to decide on filenames

    Parameters
    ----------
    files : list of str
      List of files to consider
    grouping : {'studyUID', 'accession_number', 'all', 'custom'}
      How to group DICOMs for conversion. If 'custom', see `custom_grouping`
      parameter.
    file_filter : callable, optional
      Applied to each item of filenames. Should return True if file needs to be
      kept, False otherwise.
    dcmfilter : callable, optional
      If called on dcm_data and returns True, it is used to set series_id
    flatten : bool, optional
      Creates a flattened `seqinfo` with corresponding DICOM files. True when
      invoked with `dicom_dir_template`.
    custom_grouping: str or callable, optional
      grouping key defined within heuristic. Can be a string of a
      DICOM attribute, or a method that handles more complex groupings.
    custom_seqinfo: callable, optional
      A callable which will be provided MosaicWrapper giving possibility to
      extract any custom DICOM metadata of interest.

    Returns
    -------
    seqinfo : list of list
      `seqinfo` is a list of info entries per each sequence (some entry
      there defines a key for `filegrp`)
    filegrp : dict
      `filegrp` is a dictionary with files grouped per each sequence
    """
    allowed_groupings = ["studyUID", "accession_number", "all", "custom"]
    if grouping not in allowed_groupings:
        raise ValueError("I do not know how to group by {0}".format(grouping))
    per_studyUID = grouping == "studyUID"
    # per_accession_number = grouping == 'accession_number'
    lgr.info("Analyzing %d dicoms", len(files))

    group_keys: list[SeriesID] = []
    group_values: list[int] = []
    mwgroup: list[dw.Wrapper] = []
    studyUID: Optional[str] = None

    if file_filter:
        nfl_before = len(files)
        files = list(filter(file_filter, files))
        nfl_after = len(files)
        lgr.info(
            "Filtering out {0} dicoms based on their filename".format(
                nfl_before - nfl_after
            )
        )

    if grouping == "custom":
        if custom_grouping is None:
            raise RuntimeError("Custom grouping is not defined in heuristic")
        if callable(custom_grouping):
            return custom_grouping(files, dcmfilter, SeqInfo)
        grouping = custom_grouping
        study_customgroup = None

    removeidx = []
    for idx, filename in enumerate(files):
        mwinfo = validate_dicom(filename, dcmfilter)
        if mwinfo is None:
            removeidx.append(idx)
            continue
        mw, series_id_, file_studyUID = mwinfo
        series_id = SeriesID(series_id_[0], series_id_[1])
        if per_studyUID:
            series_id = series_id._replace(file_studyUID=file_studyUID)

        if flatten:
            if per_studyUID:
                if studyUID is None:
                    studyUID = file_studyUID
                assert (
                    studyUID == file_studyUID
                ), "Conflicting study identifiers found [{}, {}].".format(
                    studyUID, file_studyUID
                )
            elif custom_grouping:
                file_customgroup = mw.dcm_data.get(grouping)
                if study_customgroup is None:
                    study_customgroup = file_customgroup
                assert (
                    study_customgroup == file_customgroup
                ), "Conflicting {0} found: [{1}, {2}]".format(
                    grouping, study_customgroup, file_customgroup
                )

        ingrp = False
        # check if same series was already converted
        for idx in range(len(mwgroup)):
            if mw.is_same_series(mwgroup[idx]):
                if grouping != "all":
                    assert (
                        mwgroup[idx].dcm_data.get("StudyInstanceUID") == file_studyUID
                    ), "Same series found for multiple different studies"
                ingrp = True
                series_id = SeriesID(
                    mwgroup[idx].dcm_data.SeriesNumber,
                    mwgroup[idx].dcm_data.ProtocolName,
                )
                if per_studyUID:
                    series_id = series_id._replace(file_studyUID=file_studyUID)
                group_keys.append(series_id)
                group_values.append(idx)

        if not ingrp:
            mwgroup.append(mw)
            group_keys.append(series_id)
            group_values.append(len(mwgroup) - 1)

    group_map = dict(zip(group_keys, group_values))

    if removeidx:
        # remove non DICOMS from files
        for idx in sorted(removeidx, reverse=True):
            del files[idx]

    seqinfos: dict[Optional[str], dict[SeqInfo, list[str]]] = {}
    flat_seqinfos: dict[SeqInfo, list[str]] = {}
    # for the next line to make any sense the series_id needs to
    # be sortable in a way that preserves the series order
    for series_id, mwidx in sorted(group_map.items()):
        mw = mwgroup[mwidx]
        series_files = [files[i] for i, s in enumerate(group_keys) if s == series_id]
        if per_studyUID:
            studyUID = series_id.file_studyUID
            series_id = series_id._replace(file_studyUID=None)
        series_id_str = str(series_id)
        if mw.image_shape is None:
            # this whole thing has no image data (maybe just PSg DICOMs)
            # If this is a Siemens PhoenixZipReport or PhysioLog, keep it:
            if mw.dcm_data.get("SeriesDescription") == "PhoenixZIPReport":
                # give it a dummy shape, so that we can continue:
                mw.image_shape = (0, 0, 0)
            else:
                # nothing to see here, just move on
                continue
        seqinfo = create_seqinfo(mw, series_files, series_id_str, custom_seqinfo)

        key: Optional[str]
        if per_studyUID:
            key = studyUID
        elif grouping == "accession_number":
            key = mw.dcm_data.get("AccessionNumber")
        elif grouping == "all":
            key = "all"
        elif custom_grouping:
            key = mw.dcm_data.get(custom_grouping)
        else:
            key = ""
        lgr.debug(
            "%30s %30s %27s %27s %5s nref=%-2d nsrc=%-2d %s"
            % (
                key,
                seqinfo.series_id,
                seqinfo.series_description,
                mw.dcm_data.ProtocolName,
                seqinfo.is_derived,
                len(mw.dcm_data.get("ReferencedImageSequence", "")),
                len(mw.dcm_data.get("SourceImageSequence", "")),
                seqinfo.image_type,
            )
        )

        if not flatten:
            seqinfos.setdefault(key, {})[seqinfo] = series_files
        else:
            flat_seqinfos[seqinfo] = series_files

    if not flatten:
        entries = len(seqinfos)
        subentries = sum(map(len, seqinfos.values()))
    else:
        entries = len(flat_seqinfos)
        subentries = sum(map(len, flat_seqinfos.values()))

    if per_studyUID:
        lgr.info(
            "Generated sequence info for %d studies with %d entries total",
            entries,
            subentries,
        )
    elif grouping == "accession_number":
        lgr.info(
            "Generated sequence info for %d accession numbers with %d entries total",
            entries,
            subentries,
        )
    else:
        lgr.info("Generated sequence info with %d entries", entries)
    if not flatten:
        return seqinfos
    else:
        return flat_seqinfos


def get_reproducible_int(dicom_list: list[str]) -> int:
    """Get integer that can be used to reproducibly sort input DICOMs, which is based on when they were acquired.

    Parameters
    ----------
    dicom_list : list[str]
        Paths to existing DICOM files

    Returns
    -------
    int
        An integer relating to when the DICOM was acquired

    Raises
    ------
    AssertionError

    Notes
    -----

    1. When date and time for can be read (see :func:`get_datetime_from_dcm`), return
        that value as time in seconds since epoch (i.e., Jan 1 1970).
    2. In cases where a date/time/datetime is not available (e.g., anonymization stripped this info), return
        epoch + AcquisitionNumber (in seconds), which is AcquisitionNumber as an integer
    3. If 1 and 2 are not possible, then raise AssertionError and provide message about missing information

    Cases are based on only the first element of the dicom_list.

    """
    import calendar

    dicom = dcm.dcmread(dicom_list[0], stop_before_pixels=True, force=True)
    dicom_datetime = get_datetime_from_dcm(dicom)
    if dicom_datetime:
        return calendar.timegm(dicom_datetime.timetuple())

    acquisition_number = dicom.get("AcquisitionNumber")
    if acquisition_number:
        return int(acquisition_number)

    raise AssertionError(
        "No metadata found that can be used to sort DICOMs reproducibly. Was header information erased?"
    )


def get_datetime_from_dcm(dcm_data: dcm.FileDataset) -> Optional[datetime.datetime]:
    """Extract datetime from filedataset, or return None is no datetime information found.

    Parameters
    ----------
    dcm_data : dcm.FileDataset
        DICOM with header, e.g., as ready by pydicom.dcmread.
        Objects with __getitem__ and have those keys with values properly formatted may also work

    Returns
    -------
    Optional[datetime.datetime]
        One of several datetimes that are related to when the scan occurred, or None if no datetime can be found

    Notes
    ------
    The following fields are checked in order

    1. AcquisitionDate & AcquisitionTime  (0008,0022); (0008,0032)
    2. AcquisitionDateTime (0008,002A);
    3. SeriesDate & SeriesTime  (0008,0021); (0008,0031)

    """
    acq_date = dcm_data.get("AcquisitionDate", "").strip()
    acq_time = dcm_data.get("AcquisitionTime", "").strip()
    if acq_date and acq_time:
        return strptime_micr(acq_date + acq_time, "%Y%m%d%H%M%S[.%f]")

    acq_dt = dcm_data.get("AcquisitionDateTime", "").strip()
    if acq_dt:
        return strptime_micr(acq_dt, "%Y%m%d%H%M%S[.%f]")

    series_date = dcm_data.get("SeriesDate", "").strip()
    series_time = dcm_data.get("SeriesTime", "").strip()
    if series_date and series_time:
        return strptime_micr(series_date + series_time, "%Y%m%d%H%M%S[.%f]")
    return None


def compress_dicoms(
    dicom_list: list[str], out_prefix: str, tempdirs: TempDirs, overwrite: bool
) -> Optional[str]:
    """Archives DICOMs into a tarball

    Also tries to do it reproducibly, so takes the date for files
    and target tarball based on the series time (within the first file)

    Parameters
    ----------
    dicom_list : list of str
      list of dicom files
    out_prefix : str
      output path prefix, including the portion of the output file name
      before .dicom.tgz suffix
    tempdirs : TempDirs
      TempDirs object to handle multiple tmpdirs
    overwrite : bool
      Overwrite existing tarfiles

    Returns
    -------
    filename : str
      Result tarball
    """

    tmpdir = tempdirs(prefix="dicomtar")
    outtar = out_prefix + ".dicom.tgz"

    if op.exists(outtar) and not overwrite:
        lgr.info("File {} already exists, will not overwrite".format(outtar))
        return None
    # tarfile encodes current time.time inside making those non-reproducible
    # so we should choose which date to use.
    # Solution from DataLad although ugly enough:

    dicom_list = sorted(dicom_list)
    dcm_time = get_reproducible_int(dicom_list)

    def _assign_dicom_time(ti: tarfile.TarInfo) -> tarfile.TarInfo:
        # Reset the date to match the one from the dicom, not from the
        # filesystem so we could sort reproducibly
        ti.mtime = dcm_time
        return ti

    with patch("time.time", lambda: dcm_time):
        try:
            if op.lexists(outtar):
                os.unlink(outtar)
            with tarfile.open(
                outtar, "w:gz", compresslevel=compresslevel, dereference=True
            ) as tar:
                for filename in dicom_list:
                    outfile = op.join(tmpdir, op.basename(filename))
                    if not op.islink(outfile):
                        os.symlink(op.realpath(filename), outfile)
                    # place into archive stripping any lead directories and
                    # adding the one corresponding to prefix
                    tar.add(
                        outfile,
                        arcname=op.join(op.basename(out_prefix), op.basename(outfile)),
                        recursive=False,
                        filter=_assign_dicom_time,
                    )
        finally:
            tempdirs.rmtree(tmpdir)

    return outtar


# Note: This function is passed to nipype by `embed_metadata_from_dicoms()`,
# and nipype reparses the function source in a clean namespace that does not
# have `from __future__ import annotations` enabled.  Thus, we need to use
# Python 3.7-compatible annotations on this function, and any non-builtin types
# used in the annotations need to be included by import statements passed to
# the `nipype.Function` constructor.
def embed_dicom_and_nifti_metadata(
    dcmfiles: List[str],
    niftifile: str,
    infofile: Union[str, Path],
    bids_info: Optional[Dict[str, Any]],
) -> None:
    """Embed metadata from nifti (affine etc) and dicoms into infofile (json)

    `niftifile` should exist. Its affine's orientation information is used while
    establishing new `NiftiImage` out of dicom stack and together with `bids_info`
    (if provided) is dumped into json `infofile`

    Parameters
    ----------
    dcmfiles
    niftifile
    infofile
    bids_info: dict
      Additional metadata to be embedded. `infofile` is overwritten if exists,
      so here you could pass some metadata which would overload (at the first
      level of the dict structure, no recursive fancy updates) what is obtained
      from nifti and dicoms

    """
    # These imports need to be within the body of the function so that they
    # will be available when executed by nipype:
    import json
    import os.path

    import dcmstack as ds
    import nibabel as nb

    from heudiconv.utils import save_json

    stack = ds.parse_and_stack(dcmfiles, force=True).values()
    if len(stack) > 1:
        raise ValueError("Found multiple series")
    # may be odict now - iter to be safe
    stack = next(iter(stack))

    if not os.path.exists(niftifile):
        raise NotImplementedError(
            "%s does not exist. "
            "We are not producing new nifti files here any longer. "
            "Use dcm2niix directly or .convert.nipype_convert helper ." % niftifile
        )

    orig_nii = nb.load(niftifile)
    aff = orig_nii.affine  # type: ignore[attr-defined]
    ornt = nb.orientations.io_orientation(aff)
    axcodes = nb.orientations.ornt2axcodes(ornt)
    new_nii = stack.to_nifti(voxel_order="".join(axcodes), embed_meta=True)
    meta_info_str = ds.NiftiWrapper(new_nii).meta_ext.to_json()
    meta_info = json.loads(meta_info_str)
    assert isinstance(meta_info, dict)

    if bids_info:
        meta_info.update(bids_info)

    # write to outfile
    save_json(infofile, meta_info)


def embed_metadata_from_dicoms(
    bids_options: Optional[str],
    item_dicoms: list[str],
    outname: str,
    outname_bids: str,
    prov_file: Optional[str],
    scaninfo: str,
    tempdirs: TempDirs,
    with_prov: bool,
) -> None:
    """
    Enhance sidecar information file with more information from DICOMs

    Parameters
    ----------
    bids_options
    item_dicoms
    outname
    outname_bids
    prov_file
    scaninfo
    tempdirs
    with_prov

    Returns
    -------

    """
    from nipype import Function, Node

    tmpdir = tempdirs(prefix="embedmeta")

    # We need to assure that paths are absolute if they are relative
    # <https://github.com/python/mypy/issues/9864>
    item_dicoms = list(map(op.abspath, item_dicoms))  # type: ignore[arg-type]

    embedfunc = Node(
        Function(
            input_names=[
                "dcmfiles",
                "niftifile",
                "infofile",
                "bids_info",
            ],
            function=embed_dicom_and_nifti_metadata,
            imports=[
                "from pathlib import Path",
                "from typing import Any, Dict, List, Optional, Union",
            ],
        ),
        name="embedder",
    )
    embedfunc.inputs.dcmfiles = item_dicoms
    embedfunc.inputs.niftifile = op.abspath(outname)
    embedfunc.inputs.infofile = op.abspath(scaninfo)
    embedfunc.inputs.bids_info = (
        load_json(op.abspath(outname_bids)) if (bids_options is not None) else None
    )
    embedfunc.base_dir = tmpdir
    cwd = os.getcwd()

    lgr.debug(
        "Embedding into %s based on dicoms[0]=%s for nifti %s",
        scaninfo,
        item_dicoms[0],
        outname,
    )
    try:
        if op.lexists(scaninfo):
            # TODO: handle annexed file case
            if not op.islink(scaninfo):
                set_readonly(scaninfo, False)
        res = embedfunc.run()
        set_readonly(scaninfo)
        if with_prov:
            assert isinstance(prov_file, str)
            g = res.provenance.rdf()
            g.parse(prov_file, format="turtle")
            g.serialize(prov_file, format="turtle")
            set_readonly(prov_file)
    except Exception as exc:
        lgr.error("Embedding failed: %s", str(exc))
        os.chdir(cwd)


def parse_private_csa_header(
    dcm_data: dcm.dataset.Dataset,
    _public_attr: str,
    private_attr: str,
    default: Optional[str] = None,
) -> str:
    """
    Parses CSA header in cases where value is not defined publicly

    Parameters
    ----------
    dcm_data : pydicom Dataset object
        DICOM metadata
    public_attr : string
        non-private DICOM attribute
    private_attr : string
        private DICOM attribute
    default (optional)
        default value if private_attr not found

    Returns
    -------
    val (default: empty string)
        private attribute value or default
    """
    # TODO: provide mapping to private_attr from public_attr
    import dcmstack.extract as dsextract
    from nibabel.nicom import csareader

    try:
        # TODO: test with attr besides ProtocolName
        csastr = csareader.get_csa_header(dcm_data, "series")["tags"][
            "MrPhoenixProtocol"
        ]["items"][0]
        csastr = csastr.replace("### ASCCONV BEGIN", "### ASCCONV BEGIN ### ")
        parsedhdr = dsextract.parse_phoenix_prot("MrPhoenixProtocol", csastr)
        val = parsedhdr[private_attr].replace(" ", "")
    except Exception as e:
        lgr.debug("Failed to parse CSA header: %s", str(e))
        val = default or ""
    assert isinstance(val, str)
    return val



"""Test functions in heudiconv.convert module.
"""
from __future__ import annotations

from glob import glob
import os.path as op
from pathlib import Path
from typing import Optional

import pytest

from heudiconv.bids import BIDSError
from heudiconv.cli.run import main as runner
import heudiconv.convert
from heudiconv.convert import (
    DW_IMAGE_IN_FMAP_FOLDER_WARNING,
    bvals_are_zero,
    update_complex_name,
    update_multiecho_name,
    update_uncombined_name,
)
from heudiconv.utils import load_heuristic

from .utils import TESTS_DATA_PATH


def test_update_complex_name() -> None:
    """Unit testing for heudiconv.convert.update_complex_name(), which updates
    filenames with the part field if appropriate.
    """
    # Standard name update
    base_fn = "sub-X_ses-Y_task-Z_run-01_sbref"
    metadata = {"ImageType": ["ORIGINAL", "PRIMARY", "P", "MB", "TE3", "ND", "MOSAIC"]}
    out_fn_true = "sub-X_ses-Y_task-Z_run-01_part-phase_sbref"
    out_fn_test = update_complex_name(metadata, base_fn)
    assert out_fn_test == out_fn_true

    # Catch an unsupported type and *do not* update
    base_fn = "sub-X_ses-Y_task-Z_run-01_phase"
    out_fn_test = update_complex_name(metadata, base_fn)
    assert out_fn_test == base_fn

    # Data type is missing from metadata so raise a RuntimeError
    base_fn = "sub-X_ses-Y_task-Z_run-01_sbref"
    metadata = {"ImageType": ["ORIGINAL", "PRIMARY", "MB", "TE3", "ND", "MOSAIC"]}
    with pytest.raises(RuntimeError):
        update_complex_name(metadata, base_fn)

    # Catch existing field with value (part is already in the filename)
    # that *does not match* metadata and raise Exception
    base_fn = "sub-X_ses-Y_task-Z_run-01_part-mag_sbref"
    metadata = {"ImageType": ["ORIGINAL", "PRIMARY", "P", "MB", "TE3", "ND", "MOSAIC"]}
    with pytest.raises(BIDSError):
        update_complex_name(metadata, base_fn)

    # Catch existing field with value (part is already in the filename)
    # that *does match* metadata and do not update
    base_fn = "sub-X_ses-Y_task-Z_run-01_part-phase_sbref"
    metadata = {"ImageType": ["ORIGINAL", "PRIMARY", "P", "MB", "TE3", "ND", "MOSAIC"]}
    out_fn_test = update_complex_name(metadata, base_fn)
    assert out_fn_test == base_fn


def test_update_multiecho_name() -> None:
    """Unit testing for heudiconv.convert.update_multiecho_name(), which updates
    filenames with the echo field if appropriate.
    """
    # Standard name update
    base_fn = "sub-X_ses-Y_task-Z_run-01_bold"
    metadata = {"EchoTime": 0.01, "EchoNumber": 1}
    echo_times = [0.01, 0.02, 0.03]
    out_fn_true = "sub-X_ses-Y_task-Z_run-01_echo-1_bold"
    out_fn_test = update_multiecho_name(metadata, base_fn, echo_times)
    assert out_fn_test == out_fn_true

    # EchoNumber field is missing from metadata, so use echo_times
    metadata = {"EchoTime": 0.01}
    out_fn_test = update_multiecho_name(metadata, base_fn, echo_times)
    assert out_fn_test == out_fn_true

    # Catch an unsupported type and *do not* update
    base_fn = "sub-X_ses-Y_task-Z_run-01_phasediff"
    out_fn_test = update_multiecho_name(metadata, base_fn, echo_times)
    assert out_fn_test == base_fn

    # EchoTime is missing, but use EchoNumber (which is the first thing it checks)
    base_fn = "sub-X_ses-Y_task-Z_run-01_bold"
    out_fn_true = "sub-X_ses-Y_task-Z_run-01_echo-1_bold"
    metadata = {"EchoNumber": 1}
    echo_times = [False, 0.02, 0.03]
    out_fn_test = update_multiecho_name(metadata, base_fn, echo_times)
    assert out_fn_test == out_fn_true

    # Both EchoTime and EchoNumber are missing, which raises a KeyError
    base_fn = "sub-X_ses-Y_task-Z_run-01_bold"
    metadata = {}
    echo_times = [False, 0.02, 0.03]
    with pytest.raises(KeyError):
        update_multiecho_name(metadata, base_fn, echo_times)

    # Providing echo times as something other than a list should raise a TypeError
    base_fn = "sub-X_ses-Y_task-Z_run-01_bold"
    with pytest.raises(TypeError):
        update_multiecho_name(metadata, base_fn, set(echo_times))  # type: ignore[arg-type]


def test_update_uncombined_name() -> None:
    """Unit testing for heudiconv.convert.update_uncombined_name(), which updates
    filenames with the ch field if appropriate.
    """
    # Standard name update
    base_fn = "sub-X_ses-Y_task-Z_run-01_bold"
    metadata = {"CoilString": "H1"}
    channel_names = ["H1", "H2", "H3", "HEA;HEP"]
    out_fn_true = "sub-X_ses-Y_task-Z_run-01_ch-01_bold"
    out_fn_test = update_uncombined_name(metadata, base_fn, channel_names)
    assert out_fn_test == out_fn_true

    # CoilString field has no number in it, so we index the channel_names list
    metadata = {"CoilString": "HEA;HEP"}
    out_fn_true = "sub-X_ses-Y_task-Z_run-01_ch-04_bold"
    out_fn_test = update_uncombined_name(metadata, base_fn, channel_names)
    assert out_fn_test == out_fn_true

    # Extract the number from the CoilString and use that
    channel_names = ["H1", "B1", "H3", "HEA;HEP"]
    metadata = {"CoilString": "H1"}
    out_fn_true = "sub-X_ses-Y_task-Z_run-01_ch-01_bold"
    out_fn_test = update_uncombined_name(metadata, base_fn, channel_names)
    assert out_fn_test == out_fn_true

    # NOTE: Extracting the number does not protect against multiple coils with the same number
    # (but, say, different letters)
    # Note that this is still "ch-01"
    metadata = {"CoilString": "B1"}
    out_fn_true = "sub-X_ses-Y_task-Z_run-01_ch-01_bold"
    out_fn_test = update_uncombined_name(metadata, base_fn, channel_names)
    assert out_fn_test == out_fn_true

    # Providing echo times as something other than a list should raise a TypeError
    base_fn = "sub-X_ses-Y_task-Z_run-01_bold"
    with pytest.raises(TypeError):
        update_uncombined_name(metadata, base_fn, set(channel_names))  # type: ignore[arg-type]


def test_b0dwi_for_fmap(tmp_path: Path, caplog: pytest.LogCaptureFixture) -> None:
    """Make sure we raise a warning when .bvec and .bval files
    are present but the modality is not dwi.
    We check it by extracting a few DICOMs from a series with
    bvals: 5 5 1500
    """
    import logging

    caplog.set_level(logging.WARNING)
    subID = "b0dwiForFmap"
    args = [
        "-c",
        "dcm2niix",
        "-o",
        str(tmp_path),
        "-b",
        "-f",
        "test_b0dwi_for_fmap",
        "--files",
        op.join(TESTS_DATA_PATH, "b0dwiForFmap"),
        "-s",
        subID,
    ]
    runner(args)

    # assert that it raised a warning that the fmap directory will contain
    # bvec and bval files.
    expected_msg = DW_IMAGE_IN_FMAP_FOLDER_WARNING.format(
        folder=op.join(tmp_path, f"sub-{subID}", "fmap")
    )
    assert any(expected_msg in c.message for c in caplog.records)

    # check that both 'fmap' and 'dwi' directories have been extracted and they contain
    # *.bvec and a *.bval files
    for mod in ["fmap", "dwi"]:
        assert op.isdir(op.join(tmp_path, f"sub-{subID}", mod))
        for ext in ["bval", "bvec"]:
            assert glob(op.join(tmp_path, f"sub-{subID}", mod, f"sub-{subID}_*.{ext}"))


# Test two scenarios for each case:
# -study without sessions
# -study with sessions
@pytest.mark.parametrize(
    "subjects, sesID, _expected_session_folder",
    [
        (["Jason", "Bourne"], None, "sub-{sID}"),
        (["Bourne"], "Treadstone", op.join("sub-{{sID}}", "ses-{{ses}}")),
    ],
)
# Two possibilities: with or without heuristics:
@pytest.mark.parametrize(
    "heuristic",
    ["example", "reproin", None],  # heuristics/example.py, heuristics/reproin.py
)
def test_populate_intended_for(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
    capfd: pytest.CaptureFixture[str],
    subjects: list[str],
    sesID: Optional[str],
    _expected_session_folder: str,
    heuristic: Optional[str],
) -> None:
    """
    Test convert

    For now, I'm just going to test that the call to populate_intended_for is
    done with the correct argument.
    More tests can be added here.
    """

    def mock_populate_intended_for(
        session: str,
        matching_parameters: str | list[str] = "Shims",
        criterion: str = "Closest",
    ) -> None:
        """
        Pretend we run populate_intended_for, but just print out the arguments.
        """
        print("session: {}".format(session))
        print("matching_parameters: {}".format(matching_parameters))
        print("criterion: {}".format(criterion))

    # mock the "populate_intended_for":
    monkeypatch.setattr(
        heudiconv.convert, "populate_intended_for", mock_populate_intended_for
    )

    outdir = op.join(tmp_path, "foo")
    outfolder = (
        op.join(outdir, "sub-{sID}", "ses-{ses}")
        if sesID
        else op.join(outdir, "sub-{sID}")
    )
    sub_ses = "sub-{sID}" + ("_ses-{ses}" if sesID else "")

    # items are a list of tuples, with each tuple having three elements:
    #   prefix, outtypes, item_dicoms
    items: list[tuple[str, tuple[str, ...], list[str]]] = [
        (
            op.join(outfolder, "anat", sub_ses + "_T1w").format(sID=s, ses=sesID),
            ("",),
            [],
        )
        for s in subjects
    ]

    heuristic_mod = load_heuristic(heuristic) if heuristic else None
    heudiconv.convert.convert(
        items,
        converter="",
        scaninfo_suffix=".json",
        custom_callable=None,
        populate_intended_for_opts=getattr(
            heuristic_mod, "POPULATE_INTENDED_FOR_OPTS", None
        ),
        with_prov=False,
        bids_options="",
        outdir=outdir,
        min_meta=True,
        overwrite=False,
    )
    output = capfd.readouterr()
    # if the heuristic module has a 'POPULATE_INTENDED_FOR_OPTS' field, we expect
    # to get the output of the mock_populate_intended_for, otherwise, no output:
    pif_cfg = getattr(heuristic_mod, "POPULATE_INTENDED_FOR_OPTS", None)
    if pif_cfg:
        assert all(
            [
                "\n".join(
                    [
                        "session: " + outfolder.format(sID=s, ses=sesID),
                        # "ImagingVolume" is defined in heuristic file; "Shims" is the default
                        f"matching_parameters: {pif_cfg['matching_parameters']}",
                        f"criterion: {pif_cfg['criterion']}",
                    ]
                )
                in output.out
                for s in subjects
            ]
        )
    else:
        # If there was no heuristic, make sure populate_intended_for was not called
        assert not output.out


def test_bvals_are_zero() -> None:
    """Unit testing for heudiconv.convert.bvals_are_zero(),
    which checks if non-dwi bvals are all zeros and can be removed
    """
    zero_bvals = op.join(TESTS_DATA_PATH, "zeros.bval")
    non_zero_bvals = op.join(TESTS_DATA_PATH, "non_zeros.bval")

    assert bvals_are_zero(zero_bvals)
    assert not bvals_are_zero(non_zero_bvals)
    assert bvals_are_zero([zero_bvals, zero_bvals])
    assert not bvals_are_zero([non_zero_bvals, zero_bvals])



"""Heuristic demonstrating conversion of the Multi-Echo sequences.

It only cares about converting sequences which have _ME_ in their
series_description and outputs to BIDS.
"""

from __future__ import annotations

from typing import Optional

from heudiconv.utils import SeqInfo


def create_key(
    template: Optional[str],
    outtype: tuple[str, ...] = ("nii.gz",),
    annotation_classes: None = None,
) -> tuple[str, tuple[str, ...], None]:
    if template is None or not template:
        raise ValueError("Template must be a valid format string")
    return (template, outtype, annotation_classes)


def infotodict(
    seqinfo: list[SeqInfo],
) -> dict[tuple[str, tuple[str, ...], None], list[str]]:
    """Heuristic evaluator for determining which runs belong where

    allowed template fields - follow python string module:

    item: index within category
    subject: participant id
    seqitem: run number during scanning
    subindex: sub index within group
    """
    bold = create_key("sub-{subject}/func/sub-{subject}_task-test_run-{item}_bold")
    megre_mag = create_key("sub-{subject}/anat/sub-{subject}_part-mag_MEGRE")
    megre_phase = create_key("sub-{subject}/anat/sub-{subject}_part-phase_MEGRE")

    info: dict[tuple[str, tuple[str, ...], None], list[str]] = {
        bold: [],
        megre_mag: [],
        megre_phase: [],
    }
    for s in seqinfo:
        if "_ME_" in s.series_description:
            info[bold].append(s.series_id)
        if "GRE_QSM" in s.series_description:
            if s.image_type[2] == "M":
                info[megre_mag].append(s.series_id)
            elif s.image_type[2] == "P":
                info[megre_phase].append(s.series_id)
    return info



from __future__ import annotations

from pathlib import Path

import pytest

from ..dlad import mark_sensitive
from ...utils import create_tree

dl = pytest.importorskip("datalad.api")


def test_mark_sensitive(tmp_path: Path) -> None:
    ds = dl.Dataset(tmp_path).create(force=True)
    create_tree(
        str(tmp_path),
        {
            "f1": "d1",
            "f2": "d2",
            "g1": "d3",
            "g2": "d1",
        },
    )
    ds.save(".")
    mark_sensitive(ds, "f*")
    all_meta = dict(ds.repo.get_metadata("."))
    target_rec = {"distribution-restrictions": ["sensitive"]}
    # g2 since the same content
    assert not all_meta.pop("g1", None)  # nothing or empty record
    assert all_meta == {"f1": target_rec, "f2": target_rec, "g2": target_rec}


def test_mark_sensitive_subset(tmp_path: Path) -> None:
    ds = dl.Dataset(tmp_path).create(force=True)
    create_tree(
        str(tmp_path),
        {
            "f1": "d1",
            "f2": "d2",
            "g1": "d3",
            "g2": "d1",
        },
    )
    ds.save(".")
    mark_sensitive(ds, "f*", [str(tmp_path / "f1")])
    all_meta = dict(ds.repo.get_metadata("."))
    target_rec = {"distribution-restrictions": ["sensitive"]}
    # g2 since the same content
    assert not all_meta.pop("g1", None)  # nothing or empty record
    assert not all_meta.pop("f2", None)  # nothing or empty record
    assert all_meta == {"f1": target_rec, "g2": target_rec}
